# -*- coding: latin-1 -*-
#
# auto generated TopDown/TMA 1.2 description for Intel Elkhart Lake
# Please see http://ark.intel.com for more details on these CPUs.
#
# References:
# http://bit.ly/tma-ispass14
# http://halobates.de/blog/p/262
# https://sites.google.com/site/analysismethods/yasin-pubs
# https://download.01.org/perfmon/
# https://github.com/andikleen/pmu-tools/wiki/toplev-manual
#

# Helpers

print_error = lambda msg: False
version = "1.2"
base_frequency = -1.0
Memory = 0
Average_Frequency = 0.0
num_cores = 1
num_threads = 1
num_sockets = 1
use_aux = False

def handle_error(obj, msg):
    print_error(msg)
    obj.errcount += 1
    obj.val = 0
    obj.thresh = False

def handle_error_metric(obj, msg):
    print_error(msg)
    obj.errcount += 1
    obj.val = 0



# Constants


# Aux. formulas


# pipeline allocation width
def Pipeline_Width(self, EV, level):
    return 4

def CLKS(self, EV, level):
    return EV("CPU_CLK_UNHALTED.CORE", level)

def CLKS_P(self, EV, level):
    return EV("CPU_CLK_UNHALTED.CORE_P", level)

def SLOTS(self, EV, level):
    return Pipeline_Width(self, EV, level) * CLKS(self, EV, level)

# Instructions Per Cycle
def IPC(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / CLKS(self, EV, level)

# Cycles Per Instruction
def CPI(self, EV, level):
    return CLKS(self, EV, level) / EV("INST_RETIRED.ANY", level)

# Uops Per Instruction
def UPI(self, EV, level):
    return EV("UOPS_RETIRED.ALL", level) / EV("INST_RETIRED.ANY", level)

# Percentage of total non-speculative loads with a store forward or unknown store address block
def Store_Fwd_Blocks(self, EV, level):
    return 100 * EV("LD_BLOCKS.DATA_UNKNOWN", level) / EV("MEM_UOPS_RETIRED.ALL_LOADS", level)

# Percentage of total non-speculative loads with a address aliasing block
def Address_Alias_Blocks(self, EV, level):
    return 100 * EV("LD_BLOCKS.4K_ALIAS", level) / EV("MEM_UOPS_RETIRED.ALL_LOADS", level)

# Percentage of total non-speculative loads that are splits
def Load_Splits(self, EV, level):
    return 100 * EV("MEM_UOPS_RETIRED.SPLIT_LOADS", level) / EV("MEM_UOPS_RETIRED.ALL_LOADS", level)

# Instructions per Branch (lower number means higher occurrence rate)
def IpBranch(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / EV("BR_INST_RETIRED.ALL_BRANCHES", level)

# Instruction per (near) call (lower number means higher occurrence rate)
def IpCall(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / EV("BR_INST_RETIRED.CALL", level)

# Instructions per Load
def IpLoad(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / EV("MEM_UOPS_RETIRED.ALL_LOADS", level)

# Instructions per Store
def IpStore(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / EV("MEM_UOPS_RETIRED.ALL_STORES", level)

# Instructions per retired Branch Misprediction
def IpMispredict(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / EV("BR_MISP_RETIRED.ALL_BRANCHES", level)

# Instructions per retired conditional Branch Misprediction where the branch was not taken
def IpMisp_Cond_Ntaken(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / (EV("BR_MISP_RETIRED.JCC", level) - EV("BR_MISP_RETIRED.TAKEN_JCC", level))

# Instructions per retired conditional Branch Misprediction where the branch was taken
def IpMisp_Cond_Taken(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / EV("BR_MISP_RETIRED.TAKEN_JCC", level)

# Instructions per retired return Branch Misprediction
def IpMisp_Ret(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / EV("BR_MISP_RETIRED.RETURN", level)

# Instructions per retired indirect call or jump Branch Misprediction
def IpMisp_Indirect(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / EV("BR_MISP_RETIRED.NON_RETURN_IND", level)

# Instructions per Far Branch
def IpFarBranch(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / (EV("BR_INST_RETIRED.FAR_BRANCH", level) / 2 )

# Ratio of all branches which mispredict
def Branch_Mispredict_Ratio(self, EV, level):
    return EV("BR_MISP_RETIRED.ALL_BRANCHES", level) / EV("BR_INST_RETIRED.ALL_BRANCHES", level)

# Ratio between Mispredicted branches and unknown branches
def Branch_Mispredict_to_Unknown_Branch_Ratio(self, EV, level):
    return EV("BR_MISP_RETIRED.ALL_BRANCHES", level) / EV("BACLEARS.ANY", level)

# Percentage of all uops which are ucode ops
def Microcode_Uop_Ratio(self, EV, level):
    return 100 * EV("UOPS_RETIRED.MS", level) / EV("UOPS_RETIRED.ALL", level)

# Percentage of all uops which are FPDiv uops
def FPDiv_Uop_Ratio(self, EV, level):
    return 100 * EV("UOPS_RETIRED.FPDIV", level) / EV("UOPS_RETIRED.ALL", level)

# Percentage of all uops which are IDiv uops
def IDiv_Uop_Ratio(self, EV, level):
    return 100 * EV("UOPS_RETIRED.IDIV", level) / EV("UOPS_RETIRED.ALL", level)

# Percentage of all uops which are x87 uops
def X87_Uop_Ratio(self, EV, level):
    return 100 * EV("UOPS_RETIRED.X87", level) / EV("UOPS_RETIRED.ALL", level)

# Average Frequency Utilization relative nominal frequency
def Turbo_Utilization(self, EV, level):
    return CLKS(self, EV, level) / EV("CPU_CLK_UNHALTED.REF_TSC", level)

# Fraction of cycles spent in Kernel mode
def Kernel_Utilization(self, EV, level):
    return EV("CPU_CLK_UNHALTED.CORE_P:sup", level) / EV("CPU_CLK_UNHALTED.CORE_P", level)

# Average CPU Utilization
def CPU_Utilization(self, EV, level):
    return EV("CPU_CLK_UNHALTED.REF_TSC", level) / EV("msr/tsc/", 0)

# Cycle cost per L2 hit
def Cycles_per_Demand_Load_L2_Hit(self, EV, level):
    return EV("MEM_BOUND_STALLS.LOAD_L2_HIT", level) / EV("MEM_LOAD_UOPS_RETIRED.L2_HIT", level)

# Cycle cost per LLC hit
def Cycles_per_Demand_Load_L3_Hit(self, EV, level):
    return EV("MEM_BOUND_STALLS.LOAD_LLC_HIT", level) / EV("MEM_LOAD_UOPS_RETIRED.L3_HIT", level)

# Cycle cost per DRAM hit
def Cycles_per_Demand_Load_DRAM_Hit(self, EV, level):
    return EV("MEM_BOUND_STALLS.LOAD_DRAM_HIT", level) / EV("MEM_LOAD_UOPS_RETIRED.DRAM_HIT", level)

# Percent of instruction miss cost that hit in the L2
def Inst_Miss_Cost_L2Hit_Percent(self, EV, level):
    return 100 * EV("MEM_BOUND_STALLS.IFETCH_L2_HIT", level) / (EV("MEM_BOUND_STALLS.IFETCH", level))

# Percent of instruction miss cost that hit in the L3
def Inst_Miss_Cost_L3Hit_Percent(self, EV, level):
    return 100 * EV("MEM_BOUND_STALLS.IFETCH_LLC_HIT", level) / (EV("MEM_BOUND_STALLS.IFETCH", level))

# Percent of instruction miss cost that hit in DRAM
def Inst_Miss_Cost_DRAMHit_Percent(self, EV, level):
    return 100 * EV("MEM_BOUND_STALLS.IFETCH_DRAM_HIT", level) / (EV("MEM_BOUND_STALLS.IFETCH", level))

# load ops retired per 1000 instruction
def MemLoadPKI(self, EV, level):
    return 1000 * EV("MEM_UOPS_RETIRED.ALL_LOADS", level) / EV("INST_RETIRED.ANY", level)

# Event groups


class Frontend_Bound:
    name = "Frontend_Bound"
    domain = "Slots"
    area = "FE"
    level = 1
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_FE_BOUND.ALL", 1) / SLOTS(self, EV, 1)
            self.thresh = (self.val > 0.20)
        except ZeroDivisionError:
            handle_error(self, "Frontend_Bound zero division")
        return self.val
    desc = """
Counts the number of issue slots  that were not consumed by
the backend due to frontend stalls."""


class Fetch_Latency:
    name = "Fetch_Latency"
    domain = "Slots"
    area = "FE"
    level = 2
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = (EV("TOPDOWN_FE_BOUND.ITLB", 2) + EV("TOPDOWN_FE_BOUND.ICACHE", 2) + EV("TOPDOWN_FE_BOUND.BRANCH_DETECT", 2) + EV("TOPDOWN_FE_BOUND.BRANCH_RESTEER", 2)) / SLOTS(self, EV, 2)
            self.thresh = (self.val > 0.15)
        except ZeroDivisionError:
            handle_error(self, "Fetch_Latency zero division")
        return self.val
    desc = """
Counts the number of issue slots  that were not delivered by
the frontend due to frontend bandwidth restrictions due to
decode, predecode, cisc, and other limitations."""


class ICache_Misses:
    name = "ICache_Misses"
    domain = "Slots"
    area = "FE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_FE_BOUND.ICACHE", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.05)
        except ZeroDivisionError:
            handle_error(self, "ICache_Misses zero division")
        return self.val
    desc = """
Counts the number of issue slots  that were not delivered by
the frontend due to instruction cache misses."""


class ITLB_Misses:
    name = "ITLB_Misses"
    domain = "Slots"
    area = "FE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_FE_BOUND.ITLB", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.05)
        except ZeroDivisionError:
            handle_error(self, "ITLB_Misses zero division")
        return self.val
    desc = """
Counts the number of issue slots  that were not delivered by
the frontend due to Instruction Table Lookaside Buffer
(ITLB) misses."""


class Branch_Detect:
    name = "Branch_Detect"
    domain = "Slots"
    area = "FE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_FE_BOUND.BRANCH_DETECT", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.05)
        except ZeroDivisionError:
            handle_error(self, "Branch_Detect zero division")
        return self.val
    desc = """
Counts the number of issue slots  that were not delivered by
the frontend due to BACLEARS, which occurs when the Branch
Target Buffer (BTB) prediction or lack thereof, was
corrected by a later branch predictor in the frontend.
Includes BACLEARS due to all branch types including
conditional and unconditional jumps, returns, and indirect
branches."""


class Branch_Resteer:
    name = "Branch_Resteer"
    domain = "Slots"
    area = "FE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_FE_BOUND.BRANCH_RESTEER", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.05)
        except ZeroDivisionError:
            handle_error(self, "Branch_Resteer zero division")
        return self.val
    desc = """
Counts the number of issue slots  that were not delivered by
the frontend due to BTCLEARS, which occurs when the Branch
Target Buffer (BTB) predicts a taken branch."""


class Fetch_Bandwidth:
    name = "Fetch_Bandwidth"
    domain = "Slots"
    area = "FE"
    level = 2
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = (EV("TOPDOWN_FE_BOUND.CISC", 2) + EV("TOPDOWN_FE_BOUND.DECODE", 2) + EV("TOPDOWN_FE_BOUND.PREDECODE", 2) + EV("TOPDOWN_FE_BOUND.OTHER", 2)) / SLOTS(self, EV, 2)
            self.thresh = (self.val > 0.10)
        except ZeroDivisionError:
            handle_error(self, "Fetch_Bandwidth zero division")
        return self.val
    desc = """
Counts the number of issue slots  that were not delivered by
the frontend due to frontend bandwidth restrictions due to
decode, predecode, cisc, and other limitations."""


class Cisc:
    name = "Cisc"
    domain = "Slots"
    area = "FE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_FE_BOUND.CISC", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.05)
        except ZeroDivisionError:
            handle_error(self, "Cisc zero division")
        return self.val
    desc = """
Counts the number of issue slots  that were not delivered by
the frontend due to the microcode sequencer (MS)."""


class Decode:
    name = "Decode"
    domain = "Slots"
    area = "FE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_FE_BOUND.DECODE", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.05)
        except ZeroDivisionError:
            handle_error(self, "Decode zero division")
        return self.val
    desc = """
Counts the number of issue slots  that were not delivered by
the frontend due to decode stalls."""


class Predecode:
    name = "Predecode"
    domain = "Slots"
    area = "FE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_FE_BOUND.PREDECODE", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.05)
        except ZeroDivisionError:
            handle_error(self, "Predecode zero division")
        return self.val
    desc = """
Counts the number of issue slots  that were not delivered by
the frontend due to wrong predecodes."""


class Other_FB:
    name = "Other_FB"
    domain = "Slots"
    area = "FE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_FE_BOUND.OTHER", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.05)
        except ZeroDivisionError:
            handle_error(self, "Other_FB zero division")
        return self.val
    desc = """
Counts the number of issue slots  that were not delivered by
the frontend due to other common frontend stalls not
categorized."""


class Bad_Speculation:
    name = "Bad_Speculation"
    domain = "Slots"
    area = "BAD"
    level = 1
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_BAD_SPECULATION.ALL", 1) / SLOTS(self, EV, 1)
            self.thresh = (self.val > 0.15)
        except ZeroDivisionError:
            handle_error(self, "Bad_Speculation zero division")
        return self.val
    desc = """
Counts the total number of issue slots that were not
consumed by the backend because allocation is stalled due to
a mispredicted jump or a machine clear. Only issue slots
wasted due to fast nukes such as memory ordering nukes are
counted. Other nukes are not accounted for. Counts all issue
slots blocked during this recovery window including relevant
microcode flows and while uops are not yet available in the
instruction queue (IQ). Also includes the issue slots that
were consumed by the backend but were thrown away because
they were younger than the mispredict or machine clear."""


class Branch_Mispredicts:
    name = "Branch_Mispredicts"
    domain = "Slots"
    area = "BAD"
    level = 2
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_BAD_SPECULATION.MISPREDICT", 2) / SLOTS(self, EV, 2)
            self.thresh = (self.val > 0.05)
        except ZeroDivisionError:
            handle_error(self, "Branch_Mispredicts zero division")
        return self.val
    desc = """
Counts the number of issue slots  that were not consumed by
the backend due to branch mispredicts."""


class Machine_Clears:
    name = "Machine_Clears"
    domain = "Slots"
    area = "BAD"
    level = 2
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = (EV("TOPDOWN_BAD_SPECULATION.MONUKE", 2)) / SLOTS(self, EV, 2)
            self.thresh = (self.val > 0.05)
        except ZeroDivisionError:
            handle_error(self, "Machine_Clears zero division")
        return self.val
    desc = """
Counts the total number of issue slots that were not
consumed by the backend because allocation is stalled due to
a machine clear (nuke) of any kind including memory ordering
and memory disambiguation."""


class Fast_Nuke:
    name = "Fast_Nuke"
    domain = "Slots"
    area = "BAD"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_BAD_SPECULATION.MONUKE", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.05)
        except ZeroDivisionError:
            handle_error(self, "Fast_Nuke zero division")
        return self.val
    desc = """
Counts the number of issue slots  that were not consumed by
the backend due to a machine clear classified as a fast nuke
due to memory ordering, memory disambiguation and memory
renaming."""


class Backend_Bound:
    name = "Backend_Bound"
    domain = "Slots"
    area = "BE"
    level = 1
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_BE_BOUND.ALL", 1) / SLOTS(self, EV, 1)
            self.thresh = (self.val > 0.10)
        except ZeroDivisionError:
            handle_error(self, "Backend_Bound zero division")
        return self.val
    desc = """
Counts the total number of issue slots  that were not
consumed by the backend due to backend stalls.  Note that
uops must be available for consumption in order for this
event to count.  If a uop is not available (IQ is empty),
this event will not count.   The rest of these subevents
count backend stalls, in cycles, due to an outstanding
request which is memory bound vs core bound.   The subevents
are not slot based events and therefore can not be precisely
added or subtracted from the Backend_Bound_Aux subevents
which are slot based."""


class Core_Bound:
    name = "Core_Bound"
    domain = "Cycles"
    area = "BE"
    level = 2
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = max(0 , self.Backend_Bound.compute(EV) - self.Memory_Bound.compute(EV))
            self.thresh = (self.val > 0.10)
        except ZeroDivisionError:
            handle_error(self, "Core_Bound zero division")
        return self.val
    desc = """
Counts the number of cycles due to backend bound stalls that
are core execution bound and not attributed to outstanding
demand load or store stalls."""


class Memory_Bound:
    name = "Memory_Bound"
    domain = "Cycles"
    area = "BE"
    level = 2
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = (EV("MEM_BOUND_STALLS.LOAD_L2_HIT", 2) + EV("MEM_BOUND_STALLS.LOAD_LLC_HIT", 2) + EV("MEM_BOUND_STALLS.LOAD_DRAM_HIT", 2) + EV("MEM_BOUND_STALLS.STORE_BUFFER_FULL", 2)) / CLKS(self, EV, 2)
            self.thresh = (self.val > 0.20)
        except ZeroDivisionError:
            handle_error(self, "Memory_Bound zero division")
        return self.val
    desc = """
Counts the number of cycles the core is stalled due to
stores or loads."""


class Store_Bound:
    name = "Store_Bound"
    domain = "Cycles"
    area = "BE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("MEM_BOUND_STALLS.STORE_BUFFER_FULL", 3) / CLKS(self, EV, 3)
            self.thresh = (self.val > 0.10)
        except ZeroDivisionError:
            handle_error(self, "Store_Bound zero division")
        return self.val
    desc = """
Counts the number of cycles the core is stalled due to store
buffer full."""


class L2_Bound:
    name = "L2_Bound"
    domain = "Cycles"
    area = "BE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("MEM_BOUND_STALLS.LOAD_L2_HIT", 3) / CLKS(self, EV, 3)
            self.thresh = (self.val > 0.10)
        except ZeroDivisionError:
            handle_error(self, "L2_Bound zero division")
        return self.val
    desc = """
Counts the number of cycles a core is stalled due to a
demand load which hit in the L2 Cache."""


class L3_Bound:
    name = "L3_Bound"
    domain = "Cycles"
    area = "BE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("MEM_BOUND_STALLS.LOAD_LLC_HIT", 3) / CLKS(self, EV, 3)
            self.thresh = (self.val > 0.10)
        except ZeroDivisionError:
            handle_error(self, "L3_Bound zero division")
        return self.val
    desc = """
Counts the number of cycles a core is stalled due to a
demand load which hit in the Last Level Cache (LLC) or other
core with HITE/F/M."""


class DRAM_Bound:
    name = "DRAM_Bound"
    domain = "Cycles"
    area = "BE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("MEM_BOUND_STALLS.LOAD_DRAM_HIT", 3) / CLKS(self, EV, 3)
            self.thresh = (self.val > 0.10)
        except ZeroDivisionError:
            handle_error(self, "DRAM_Bound zero division")
        return self.val
    desc = """
Counts the number of cycles the core is stalled due to a
demand load miss which hit in DRAM or MMIO (Non-DRAM)."""


class Backend_Bound_Aux:
    name = "Backend_Bound_Aux"
    domain = "Slots"
    area = "BE_aux"
    level = 1
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = self.Backend_Bound.compute(EV)
            self.thresh = (self.val > 0.20)
        except ZeroDivisionError:
            handle_error(self, "Backend_Bound_Aux zero division")
        return self.val
    desc = """
Counts the total number of issue slots  that were not
consumed by the backend due to backend stalls.  Note that
UOPS must be available for consumption in order for this
event to count.  If a uop is not available (IQ is empty),
this event will not count.  All of these subevents count
backend stalls, in slots, due to a resource limitation.
These are not cycle based events and therefore can not be
precisely added or subtracted from the Backend_Bound
subevents which are cycle based.  These subevents are
supplementary to Backend_Bound and can be used to analyze
results from a resource perspective at allocation."""


class Resource_Bound:
    name = "Resource_Bound"
    domain = "Slots"
    area = "BE_aux"
    level = 2
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = self.Backend_Bound.compute(EV)
            self.thresh = (self.val > 0.20)
        except ZeroDivisionError:
            handle_error(self, "Resource_Bound zero division")
        return self.val
    desc = """
Counts the total number of issue slots  that were not
consumed by the backend due to backend stalls.  Note that
uops must be available for consumption in order for this
event to count.  If a uop is not available (IQ is empty),
this event will not count."""


class Mem_Scheduler:
    name = "Mem_Scheduler"
    domain = "Slots"
    area = "BE_aux"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_BE_BOUND.MEM_SCHEDULER", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.10)
        except ZeroDivisionError:
            handle_error(self, "Mem_Scheduler zero division")
        return self.val
    desc = """
Counts the number of issue slots  that were not consumed by
the backend due to memory reservation stalls in which a
scheduler is not able to accept uops."""


class Non_Mem_Scheduler:
    name = "Non_Mem_Scheduler"
    domain = "Slots"
    area = "BE_aux"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_BE_BOUND.NON_MEM_SCHEDULER", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.10)
        except ZeroDivisionError:
            handle_error(self, "Non_Mem_Scheduler zero division")
        return self.val
    desc = """
Counts the number of issue slots  that were not consumed by
the backend due to IEC or FPC RAT stalls, which can be due
to FIQ or IEC reservation stalls in which the integer,
floating point or SIMD scheduler is not able to accept uops."""


class Register:
    name = "Register"
    domain = "Slots"
    area = "BE_aux"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_BE_BOUND.REGISTER", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.10)
        except ZeroDivisionError:
            handle_error(self, "Register zero division")
        return self.val
    desc = """
Counts the number of issue slots  that were not consumed by
the backend due to the physical register file unable to
accept an entry (marble stalls)."""


class Reorder_Buffer:
    name = "Reorder_Buffer"
    domain = "Slots"
    area = "BE_aux"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_BE_BOUND.REORDER_BUFFER", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.10)
        except ZeroDivisionError:
            handle_error(self, "Reorder_Buffer zero division")
        return self.val
    desc = """
Counts the number of issue slots  that were not consumed by
the backend due to the reorder buffer being full (ROB
stalls)."""


class Store_Buffer:
    name = "Store_Buffer"
    domain = "Slots"
    area = "BE_aux"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_BE_BOUND.STORE_BUFFER", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.10)
        except ZeroDivisionError:
            handle_error(self, "Store_Buffer zero division")
        return self.val
    desc = """
Counts the number of issue slots  that were not consumed by
the backend due to store buffers stalls."""


class Alloc_Restriction:
    name = "Alloc_Restriction"
    domain = "Slots"
    area = "BE_aux"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_BE_BOUND.ALLOC_RESTRICTIONS", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.10)
        except ZeroDivisionError:
            handle_error(self, "Alloc_Restriction zero division")
        return self.val
    desc = """
Counts the number of issue slots  that were not consumed by
the backend due to certain allocation restrictions."""


class Serialization:
    name = "Serialization"
    domain = "Slots"
    area = "BE_aux"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_BE_BOUND.SERIALIZATION", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.10)
        except ZeroDivisionError:
            handle_error(self, "Serialization zero division")
        return self.val
    desc = """
Counts the number of issue slots  that were not consumed by
the backend due to scoreboards from the instruction queue
(IQ), jump execution unit (JEU), or microcode sequencer
(MS)."""


class Retiring:
    name = "Retiring"
    domain = "Slots"
    area = "RET"
    level = 1
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_RETIRING.ALL", 1) / SLOTS(self, EV, 1)
            self.thresh = (self.val > 0.75)
        except ZeroDivisionError:
            handle_error(self, "Retiring zero division")
        return self.val
    desc = """
Counts the number of issue slots  that result in retirement
slots."""


class Base:
    name = "Base"
    domain = "Slots"
    area = "RET"
    level = 2
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = (EV("TOPDOWN_RETIRING.ALL", 2) - EV("UOPS_RETIRED.MS", 2)) / SLOTS(self, EV, 2)
            self.thresh = (self.val > 0.60)
        except ZeroDivisionError:
            handle_error(self, "Base zero division")
        return self.val
    desc = """
Counts the number of uops that are not from the
microsequencer."""


class FPDIV_uops:
    name = "FPDIV_uops"
    domain = "Slots"
    area = "RET"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("UOPS_RETIRED.FPDIV", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.20)
        except ZeroDivisionError:
            handle_error(self, "FPDIV_uops zero division")
        return self.val
    desc = """
Counts the number of floating point divide operations per
uop."""


class Other_Ret:
    name = "Other_Ret"
    domain = "Slots"
    area = "RET"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = (EV("TOPDOWN_RETIRING.ALL", 3) - EV("UOPS_RETIRED.MS", 3) - EV("UOPS_RETIRED.FPDIV", 3)) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.30)
        except ZeroDivisionError:
            handle_error(self, "Other_Ret zero division")
        return self.val
    desc = """
Counts the number of uops retired excluding ms and fp div
uops."""


class MS_uops:
    name = "MS_uops"
    domain = "Slots"
    area = "RET"
    level = 2
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("UOPS_RETIRED.MS", 2) / SLOTS(self, EV, 2)
            self.thresh = (self.val > 0.05)
        except ZeroDivisionError:
            handle_error(self, "MS_uops zero division")
        return self.val
    desc = """
Counts the number of uops that are from the complex flows
issued by the micro-sequencer (MS).  This includes uops from
flows due to complex instructions, faults, assists, and
inserted flows."""


class Metric_CLKS:
    name = "CLKS"
    domain = "Cycles"
    maxval = 0
    errcount = 0
    area = "Info.Core"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = CLKS(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "CLKS zero division")
    desc = """
"""


class Metric_CLKS_P:
    name = "CLKS_P"
    domain = "Cycles"
    maxval = 0
    errcount = 0
    area = "Info.Core"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = CLKS_P(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "CLKS_P zero division")
    desc = """
"""


class Metric_SLOTS:
    name = "SLOTS"
    domain = "Cycles"
    maxval = 0
    errcount = 0
    area = "Info.Core"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = SLOTS(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "SLOTS zero division")
    desc = """
"""


class Metric_IPC:
    name = "IPC"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Core"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IPC(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IPC zero division")
    desc = """
Instructions Per Cycle"""


class Metric_CPI:
    name = "CPI"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Core"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = CPI(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "CPI zero division")
    desc = """
Cycles Per Instruction"""


class Metric_UPI:
    name = "UPI"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Core"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = UPI(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "UPI zero division")
    desc = """
Uops Per Instruction"""


class Metric_Store_Fwd_Blocks:
    name = "Store_Fwd_Blocks"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.L1_Bound"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Store_Fwd_Blocks(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Store_Fwd_Blocks zero division")
    desc = """
Percentage of total non-speculative loads with a store
forward or unknown store address block"""


class Metric_Address_Alias_Blocks:
    name = "Address_Alias_Blocks"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.L1_Bound"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Address_Alias_Blocks(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Address_Alias_Blocks zero division")
    desc = """
Percentage of total non-speculative loads with a address
aliasing block"""


class Metric_Load_Splits:
    name = "Load_Splits"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.L1_Bound"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Load_Splits(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Load_Splits zero division")
    desc = """
Percentage of total non-speculative loads that are splits"""


class Metric_IpBranch:
    name = "IpBranch"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Inst_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpBranch(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpBranch zero division")
    desc = """
Instructions per Branch (lower number means higher
occurrence rate)"""


class Metric_IpCall:
    name = "IpCall"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Inst_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpCall(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpCall zero division")
    desc = """
Instruction per (near) call (lower number means higher
occurrence rate)"""


class Metric_IpLoad:
    name = "IpLoad"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Inst_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpLoad(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpLoad zero division")
    desc = """
Instructions per Load"""


class Metric_IpStore:
    name = "IpStore"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Inst_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpStore(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpStore zero division")
    desc = """
Instructions per Store"""


class Metric_IpMispredict:
    name = "IpMispredict"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Inst_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpMispredict(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpMispredict zero division")
    desc = """
Instructions per retired Branch Misprediction"""


class Metric_IpMisp_Cond_Ntaken:
    name = "IpMisp_Cond_Ntaken"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Inst_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpMisp_Cond_Ntaken(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpMisp_Cond_Ntaken zero division")
    desc = """
Instructions per retired conditional Branch Misprediction
where the branch was not taken"""


class Metric_IpMisp_Cond_Taken:
    name = "IpMisp_Cond_Taken"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Inst_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpMisp_Cond_Taken(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpMisp_Cond_Taken zero division")
    desc = """
Instructions per retired conditional Branch Misprediction
where the branch was taken"""


class Metric_IpMisp_Ret:
    name = "IpMisp_Ret"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Inst_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpMisp_Ret(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpMisp_Ret zero division")
    desc = """
Instructions per retired return Branch Misprediction"""


class Metric_IpMisp_Indirect:
    name = "IpMisp_Indirect"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Inst_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpMisp_Indirect(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpMisp_Indirect zero division")
    desc = """
Instructions per retired indirect call or jump Branch
Misprediction"""


class Metric_IpFarBranch:
    name = "IpFarBranch"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Inst_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpFarBranch(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpFarBranch zero division")
    desc = """
Instructions per Far Branch"""


class Metric_Branch_Mispredict_Ratio:
    name = "Branch_Mispredict_Ratio"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Inst_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Branch_Mispredict_Ratio(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Branch_Mispredict_Ratio zero division")
    desc = """
Ratio of all branches which mispredict"""


class Metric_Branch_Mispredict_to_Unknown_Branch_Ratio:
    name = "Branch_Mispredict_to_Unknown_Branch_Ratio"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Inst_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Branch_Mispredict_to_Unknown_Branch_Ratio(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Branch_Mispredict_to_Unknown_Branch_Ratio zero division")
    desc = """
Ratio between Mispredicted branches and unknown branches"""


class Metric_Microcode_Uop_Ratio:
    name = "Microcode_Uop_Ratio"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Inst_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Microcode_Uop_Ratio(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Microcode_Uop_Ratio zero division")
    desc = """
Percentage of all uops which are ucode ops"""


class Metric_FPDiv_Uop_Ratio:
    name = "FPDiv_Uop_Ratio"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Inst_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = FPDiv_Uop_Ratio(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "FPDiv_Uop_Ratio zero division")
    desc = """
Percentage of all uops which are FPDiv uops"""


class Metric_IDiv_Uop_Ratio:
    name = "IDiv_Uop_Ratio"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Inst_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IDiv_Uop_Ratio(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IDiv_Uop_Ratio zero division")
    desc = """
Percentage of all uops which are IDiv uops"""


class Metric_X87_Uop_Ratio:
    name = "X87_Uop_Ratio"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Inst_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = X87_Uop_Ratio(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "X87_Uop_Ratio zero division")
    desc = """
Percentage of all uops which are x87 uops"""


class Metric_Turbo_Utilization:
    name = "Turbo_Utilization"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.System"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Turbo_Utilization(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Turbo_Utilization zero division")
    desc = """
Average Frequency Utilization relative nominal frequency"""


class Metric_Kernel_Utilization:
    name = "Kernel_Utilization"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.System"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Kernel_Utilization(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Kernel_Utilization zero division")
    desc = """
Fraction of cycles spent in Kernel mode"""


class Metric_CPU_Utilization:
    name = "CPU_Utilization"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.System"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = CPU_Utilization(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "CPU_Utilization zero division")
    desc = """
Average CPU Utilization"""


class Metric_Cycles_per_Demand_Load_L2_Hit:
    name = "Cycles_per_Demand_Load_L2_Hit"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Memory"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Cycles_per_Demand_Load_L2_Hit(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Cycles_per_Demand_Load_L2_Hit zero division")
    desc = """
Cycle cost per L2 hit"""


class Metric_Cycles_per_Demand_Load_L3_Hit:
    name = "Cycles_per_Demand_Load_L3_Hit"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Memory"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Cycles_per_Demand_Load_L3_Hit(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Cycles_per_Demand_Load_L3_Hit zero division")
    desc = """
Cycle cost per LLC hit"""


class Metric_Cycles_per_Demand_Load_DRAM_Hit:
    name = "Cycles_per_Demand_Load_DRAM_Hit"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Memory"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Cycles_per_Demand_Load_DRAM_Hit(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Cycles_per_Demand_Load_DRAM_Hit zero division")
    desc = """
Cycle cost per DRAM hit"""


class Metric_Inst_Miss_Cost_L2Hit_Percent:
    name = "Inst_Miss_Cost_L2Hit_Percent"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Frontend"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Inst_Miss_Cost_L2Hit_Percent(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Inst_Miss_Cost_L2Hit_Percent zero division")
    desc = """
Percent of instruction miss cost that hit in the L2"""


class Metric_Inst_Miss_Cost_L3Hit_Percent:
    name = "Inst_Miss_Cost_L3Hit_Percent"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Frontend"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Inst_Miss_Cost_L3Hit_Percent(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Inst_Miss_Cost_L3Hit_Percent zero division")
    desc = """
Percent of instruction miss cost that hit in the L3"""


class Metric_Inst_Miss_Cost_DRAMHit_Percent:
    name = "Inst_Miss_Cost_DRAMHit_Percent"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Frontend"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Inst_Miss_Cost_DRAMHit_Percent(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Inst_Miss_Cost_DRAMHit_Percent zero division")
    desc = """
Percent of instruction miss cost that hit in DRAM"""


class Metric_MemLoadPKI:
    name = "MemLoadPKI"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Memory"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = MemLoadPKI(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "MemLoadPKI zero division")
    desc = """
load ops retired per 1000 instruction"""


# Schedule



class Setup:
    def __init__(self, r):
        o = dict()
        n = Frontend_Bound() ; r.run(n) ; o["Frontend_Bound"] = n
        n = Fetch_Latency() ; r.run(n) ; o["Fetch_Latency"] = n
        n = ICache_Misses() ; r.run(n) ; o["ICache_Misses"] = n
        n = ITLB_Misses() ; r.run(n) ; o["ITLB_Misses"] = n
        n = Branch_Detect() ; r.run(n) ; o["Branch_Detect"] = n
        n = Branch_Resteer() ; r.run(n) ; o["Branch_Resteer"] = n
        n = Fetch_Bandwidth() ; r.run(n) ; o["Fetch_Bandwidth"] = n
        n = Cisc() ; r.run(n) ; o["Cisc"] = n
        n = Decode() ; r.run(n) ; o["Decode"] = n
        n = Predecode() ; r.run(n) ; o["Predecode"] = n
        n = Other_FB() ; r.run(n) ; o["Other_FB"] = n
        n = Bad_Speculation() ; r.run(n) ; o["Bad_Speculation"] = n
        n = Branch_Mispredicts() ; r.run(n) ; o["Branch_Mispredicts"] = n
        n = Machine_Clears() ; r.run(n) ; o["Machine_Clears"] = n
        n = Fast_Nuke() ; r.run(n) ; o["Fast_Nuke"] = n
        n = Backend_Bound() ; r.run(n) ; o["Backend_Bound"] = n
        n = Core_Bound() ; r.run(n) ; o["Core_Bound"] = n
        n = Memory_Bound() ; r.run(n) ; o["Memory_Bound"] = n
        n = Store_Bound() ; r.run(n) ; o["Store_Bound"] = n
        n = L2_Bound() ; r.run(n) ; o["L2_Bound"] = n
        n = L3_Bound() ; r.run(n) ; o["L3_Bound"] = n
        n = DRAM_Bound() ; r.run(n) ; o["DRAM_Bound"] = n
        if use_aux:
            n = Backend_Bound_Aux() ; r.run(n) ; o["Backend_Bound_Aux"] = n
        if use_aux:
            n = Resource_Bound() ; r.run(n) ; o["Resource_Bound"] = n
        if use_aux:
            n = Mem_Scheduler() ; r.run(n) ; o["Mem_Scheduler"] = n
        if use_aux:
            n = Non_Mem_Scheduler() ; r.run(n) ; o["Non_Mem_Scheduler"] = n
        if use_aux:
            n = Register() ; r.run(n) ; o["Register"] = n
        if use_aux:
            n = Reorder_Buffer() ; r.run(n) ; o["Reorder_Buffer"] = n
        if use_aux:
            n = Store_Buffer() ; r.run(n) ; o["Store_Buffer"] = n
        if use_aux:
            n = Alloc_Restriction() ; r.run(n) ; o["Alloc_Restriction"] = n
        if use_aux:
            n = Serialization() ; r.run(n) ; o["Serialization"] = n
        n = Retiring() ; r.run(n) ; o["Retiring"] = n
        n = Base() ; r.run(n) ; o["Base"] = n
        n = FPDIV_uops() ; r.run(n) ; o["FPDIV_uops"] = n
        n = Other_Ret() ; r.run(n) ; o["Other_Ret"] = n
        n = MS_uops() ; r.run(n) ; o["MS_uops"] = n

        # parents

        o["Fetch_Latency"].parent = o["Frontend_Bound"]
        o["ICache_Misses"].parent = o["Fetch_Latency"]
        o["ITLB_Misses"].parent = o["Fetch_Latency"]
        o["Branch_Detect"].parent = o["Fetch_Latency"]
        o["Branch_Resteer"].parent = o["Fetch_Latency"]
        o["Fetch_Bandwidth"].parent = o["Frontend_Bound"]
        o["Cisc"].parent = o["Fetch_Bandwidth"]
        o["Decode"].parent = o["Fetch_Bandwidth"]
        o["Predecode"].parent = o["Fetch_Bandwidth"]
        o["Other_FB"].parent = o["Fetch_Bandwidth"]
        o["Branch_Mispredicts"].parent = o["Bad_Speculation"]
        o["Machine_Clears"].parent = o["Bad_Speculation"]
        o["Fast_Nuke"].parent = o["Machine_Clears"]
        o["Core_Bound"].parent = o["Backend_Bound"]
        o["Memory_Bound"].parent = o["Backend_Bound"]
        o["Store_Bound"].parent = o["Memory_Bound"]
        o["L2_Bound"].parent = o["Memory_Bound"]
        o["L3_Bound"].parent = o["Memory_Bound"]
        o["DRAM_Bound"].parent = o["Memory_Bound"]
        if use_aux:
            o["Resource_Bound"].parent = o["Backend_Bound_Aux"]
        if use_aux:
            o["Mem_Scheduler"].parent = o["Resource_Bound"]
        if use_aux:
            o["Non_Mem_Scheduler"].parent = o["Resource_Bound"]
        if use_aux:
            o["Register"].parent = o["Resource_Bound"]
        if use_aux:
            o["Reorder_Buffer"].parent = o["Resource_Bound"]
        if use_aux:
            o["Store_Buffer"].parent = o["Resource_Bound"]
        if use_aux:
            o["Alloc_Restriction"].parent = o["Resource_Bound"]
        if use_aux:
            o["Serialization"].parent = o["Resource_Bound"]
        o["Base"].parent = o["Retiring"]
        o["FPDIV_uops"].parent = o["Base"]
        o["Other_Ret"].parent = o["Base"]
        o["MS_uops"].parent = o["Retiring"]

        # user visible metrics

        n = Metric_CLKS() ; r.metric(n) ; o["CLKS"] = n
        n = Metric_CLKS_P() ; r.metric(n) ; o["CLKS_P"] = n
        n = Metric_SLOTS() ; r.metric(n) ; o["SLOTS"] = n
        n = Metric_IPC() ; r.metric(n) ; o["IPC"] = n
        n = Metric_CPI() ; r.metric(n) ; o["CPI"] = n
        n = Metric_UPI() ; r.metric(n) ; o["UPI"] = n
        n = Metric_Store_Fwd_Blocks() ; r.metric(n) ; o["Store_Fwd_Blocks"] = n
        n = Metric_Address_Alias_Blocks() ; r.metric(n) ; o["Address_Alias_Blocks"] = n
        n = Metric_Load_Splits() ; r.metric(n) ; o["Load_Splits"] = n
        n = Metric_IpBranch() ; r.metric(n) ; o["IpBranch"] = n
        n = Metric_IpCall() ; r.metric(n) ; o["IpCall"] = n
        n = Metric_IpLoad() ; r.metric(n) ; o["IpLoad"] = n
        n = Metric_IpStore() ; r.metric(n) ; o["IpStore"] = n
        n = Metric_IpMispredict() ; r.metric(n) ; o["IpMispredict"] = n
        n = Metric_IpMisp_Cond_Ntaken() ; r.metric(n) ; o["IpMisp_Cond_Ntaken"] = n
        n = Metric_IpMisp_Cond_Taken() ; r.metric(n) ; o["IpMisp_Cond_Taken"] = n
        n = Metric_IpMisp_Ret() ; r.metric(n) ; o["IpMisp_Ret"] = n
        n = Metric_IpMisp_Indirect() ; r.metric(n) ; o["IpMisp_Indirect"] = n
        n = Metric_IpFarBranch() ; r.metric(n) ; o["IpFarBranch"] = n
        n = Metric_Branch_Mispredict_Ratio() ; r.metric(n) ; o["Branch_Mispredict_Ratio"] = n
        n = Metric_Branch_Mispredict_to_Unknown_Branch_Ratio() ; r.metric(n) ; o["Branch_Mispredict_to_Unknown_Branch_Ratio"] = n
        n = Metric_Microcode_Uop_Ratio() ; r.metric(n) ; o["Microcode_Uop_Ratio"] = n
        n = Metric_FPDiv_Uop_Ratio() ; r.metric(n) ; o["FPDiv_Uop_Ratio"] = n
        n = Metric_IDiv_Uop_Ratio() ; r.metric(n) ; o["IDiv_Uop_Ratio"] = n
        n = Metric_X87_Uop_Ratio() ; r.metric(n) ; o["X87_Uop_Ratio"] = n
        n = Metric_Turbo_Utilization() ; r.metric(n) ; o["Turbo_Utilization"] = n
        n = Metric_Kernel_Utilization() ; r.metric(n) ; o["Kernel_Utilization"] = n
        n = Metric_CPU_Utilization() ; r.metric(n) ; o["CPU_Utilization"] = n
        n = Metric_Cycles_per_Demand_Load_L2_Hit() ; r.metric(n) ; o["Cycles_per_Demand_Load_L2_Hit"] = n
        n = Metric_Cycles_per_Demand_Load_L3_Hit() ; r.metric(n) ; o["Cycles_per_Demand_Load_L3_Hit"] = n
        n = Metric_Cycles_per_Demand_Load_DRAM_Hit() ; r.metric(n) ; o["Cycles_per_Demand_Load_DRAM_Hit"] = n
        n = Metric_Inst_Miss_Cost_L2Hit_Percent() ; r.metric(n) ; o["Inst_Miss_Cost_L2Hit_Percent"] = n
        n = Metric_Inst_Miss_Cost_L3Hit_Percent() ; r.metric(n) ; o["Inst_Miss_Cost_L3Hit_Percent"] = n
        n = Metric_Inst_Miss_Cost_DRAMHit_Percent() ; r.metric(n) ; o["Inst_Miss_Cost_DRAMHit_Percent"] = n
        n = Metric_MemLoadPKI() ; r.metric(n) ; o["MemLoadPKI"] = n

        # references between groups

        o["Core_Bound"].Memory_Bound = o["Memory_Bound"]
        o["Core_Bound"].Backend_Bound = o["Backend_Bound"]
        if use_aux:
            o["Backend_Bound_Aux"].Backend_Bound = o["Backend_Bound"]
        if use_aux:
            o["Resource_Bound"].Backend_Bound = o["Backend_Bound"]

        # siblings cross-tree




# track mmap updates in a perf stream and allow lookup of symbols
#
# Copyright (c) 2013-2014, Intel Corporation
# Author: Andi Kleen
#
# This program is free software; you can redistribute it and/or modify it
# under the terms and conditions of the GNU General Public License,
# version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.

from collections import defaultdict
import bisect

# max reorder window for MMAP updates
LOOKAHEAD_WINDOW = 1024

def lookup(m, ip):
    i = bisect.bisect_left(m, (ip,))
    if i < len(m) and m[i][0] == ip:
        mr = m[i]
    elif i == 0:
        return None, 0
    else:
        mr = m[i - 1]
    return mr, ip - mr[0]

class MmapTracker:
    """Track mmap updates in a perf stream and allow lookup of symbols."""

    def __init__(self):
        self.maps = defaultdict(list)
        self.pnames = defaultdict(str)
        self.lookahead = 0
        self.updates = []

    # look ahead for out of order mmap updates
    def lookahead_mmap(self, ev, n):
        if n - self.lookahead == 0:
            self.lookahead = min(n + LOOKAHEAD_WINDOW, len(ev))
            for l in range(n, self.lookahead):
                j = ev[l]
                # no time stamp: assume it's synthesized and kernel
                if j.type in ('MMAP','MMAP2') and j.pid == -1 and j.tid == 0:
                    bisect.insort(self.maps[j.pid],
                                  (j.addr, j.len, j.filename))
                elif j.type in ('COMM','MMAP','MMAP2'):
                    bisect.insort(self.updates, (j.time2, j))

    # process pending updates for a sample
    def update_sample(self, j):
        updates = self.updates
        while len(updates) > 0 and j.time >= updates[0][0]:
            u = updates[0][1]
            del updates[0]
            if u.type in ('MMAP', 'MMAP2'):
                pid = u.pid
                bisect.insort(self.maps[pid], (u.addr, u.len, u.filename))
            elif u.type == 'COMM':
                self.maps[u.pid] = []
                self.pnames[u.pid] = u.comm

    # look up tables with current state
    def resolve(self, pid, ip):
        if not self.maps[pid]:
            # xxx kernel
            return None, None, 0
        m, offset = lookup(self.maps[pid], ip)
        if not m or offset >= m[1]:
            # look up kernel
            m, offset = lookup(self.maps[-1], ip)
            if not m or offset >= m[1]:
                return None, None, 0
        assert m[0] <= ip <= m[0] + m[1]
        return m[2], m[0], offset



#!/usr/bin/env python3
# configure latego workaround on Sandy Bridge EP
# can be run as a standalone tool or used as module
# latego enable|disable hexevent
from __future__ import print_function
import signal
import struct
import re
import os
import msr
import pci

busses = (0x3f, 0x7f, 0xbf, 0xff)

def local_direct2core(val):
    c = 0
    for b in busses:
        if pci.probe(b, 14, 0):
            pci.changebit(b, 14, 0, 0x84, 1, val)
            c += 1
    if c == 0:
        print("no local devices found")

def remote_direct2core(val):
    c = 0
    for b in busses:
        if pci.probe(b, 8, 0):
            pci.changebit(b, 8, 0, 0x80, 1, val)
            pci.changebit(b, 9, 0, 0x80, 1, val)
            c += 1
    if c == 0:
        print("no remote devices found")

def direct2core(val):
    # make sure all cores are awake when we do that
    f = os.open("/dev/cpu_dma_latency", os.O_WRONLY)
    os.write(f, struct.pack("I", 0))
    local_direct2core(val)
    remote_direct2core(val)
    os.close(f)

def set_bypass(val):
    msr.changebit(0x39c, 0, val)

bypass  = 1 << 0
d2c     = 1 << 1
latego_events = {
    0x04d1: bypass,
    0x20d1: bypass|d2c,
    0x01d3: bypass|d2c,
    0x04d3: bypass|d2c,
    0x01d2: bypass,
    0x02d2: bypass,
    0x04d2: bypass,
    0x08d2: bypass,
    0x01cd: bypass|d2c
}

latego_names = {
    "mem_load_uops_retired.llc_hit": 0x04d1,
    "mem_load_uops_retired.llc_miss": 0x20d1,
    "mem_load_uops_llc_miss_retired.local_dram": 0x01d3,
    "mem_load_uops_llc_miss_retired.remote_dram": 0x04d3,
    "mem_load_uops_llc_hit_retired.xsnp_miss": 0x01d2,
    "mem_load_uops_llc_hit_retired.xsnp_hit": 0x02d2,
    "mem_load_uops_llc_hit_retired.xsnp_hitm": 0x04d2,
    "mem_load_uops_llc_hit_retired.xsnp_none": 0x08d2,
    "mem_trans_retired.load_latency": 0x01cd
}

signal_setup = False
enabled = 0

def cleanup():
    if enabled & bypass:
        set_bypass(0)
    if enabled & d2c:
        direct2core(0)

def get_event(e):
    if re.match(r"[0-9]+", e):
        return int(e, 16)
    if e in latego_names:
        return latego_names[e]
    return e

def setup_event(event, val):
    global signal_setup
    global enabled
    action = ("Disabling", "Enabling")[val]
    if val and not signal_setup:
        signal.signal(signal.SIGINT, cleanup)
        signal.signal(signal.SIGPIPE, cleanup)
        signal_setup = True
    if event in latego_events:
        v = latego_events[event]
        if v & d2c:
            print("%s direct2core" % (action))
            direct2core(val)
        if v & bypass:
            print("%s bypass" % (action))
            set_bypass(val)
        if val:
            enabled = v
        else:
            enabled = 0

if __name__ == '__main__':
    import sys
    if len(sys.argv) == 3 and sys.argv[1] == "enable":
        setup_event(get_event(sys.argv[2]), 1)
    elif len(sys.argv) == 3 and sys.argv[1] == "disable":
        setup_event(get_event(sys.argv[2]), 0)
    elif len(sys.argv) == 2 and sys.argv[1] == "list":
        print("%-45s %04s" % ("name", "hex"))
        for i in latego_names.keys():
            print("%-45s %04x" % (i, latego_names[i], ))
    else:
        print("Usage: latego enable|disable hexevent|namedevent")
        print("       latego list")
        sys.exit(1)



#!/usr/bin/env python3
# counterdiff.py < plog program ..      (or general perf arguments)
# verify plog.* output from toplev by running event one by one
# this can be used to estimate multiplexing measurement errors
from __future__ import print_function
import sys, os

def run(x):
    print(x)
    os.system(x)

for l in sys.stdin:
    if l.find(",") < 0:
        continue
    n = l.strip().split(",")
    run("perf stat --output l -x, -e %s %s" %
        (n[1], " ".join(sys.argv[1:])))
    f = open("l", "r")
    for i in f:
        if i.find(",") < 0:
            continue
        j = i.strip().split(",")
        break
    f.close()
    if float(n[0]) > 0:
        delta = (float(j[0]) - float(n[0])) / float(n[0])
    else:
        delta = 0
    print(n[1], j[0], n[0], "%.2f" % (delta * 100.0))



# Copyright (c) 2018 Technical University of Munich
# Author: Martin Becker
#
# This program is free software; you can redistribute it and/or modify it
# under the terms and conditions of the GNU General Public License,
# version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
import math
import logging
import operator
from tl_io import warn

log = logging.getLogger(__name__)
TEMPVAL = 'anon'

div_op = operator.div if 'div' in operator.__dict__ else None # type: ignore

def combine_uval(ulist):
    """
    Combine multiple measurements of the same event into one measurement.
    Uses weighted average.
    """
    combined = None
    if ulist is not None:
        combined = ulist[0]
        for oth in ulist[1:]:
            combined.update(oth)
    return combined


class UVal:
    """
    Measurement value annotated with uncertainty. Supports binary operators for error propagation.
    """

    def __init__(self, name, value, stddev=0., samples=1, mux=100., comment="", computed=False):
        self.name = name
        self.comment = comment
        self.value = value
        self.stddev = stddev
        self.samples = samples
        self.computed = computed
        self.is_ratio = False
        self.multiplex = mux

    def __repr__(self):
        return "{} [{} +- {}]*{}".format(self.name, self.value, self.stddev, self.samples)

    def format_value(self, unit):
        if self.value is None:
            return ""
        if self.is_ratio:
            return "{:>16.1f} ".format(self.value * 100.)
        elif unit == "Count" or unit == "Clocks":
            return "{:16,.0f}   ".format(self.value)
        elif self.value > 1000:
            return "{:16,.1f} ".format(self.value)
        else:
            return "{:16.2f}".format(self.value)

    def format_value_raw(self):
        if self.value is None:
            return ""
        if self.is_ratio:
            return "{:>13.1f}".format(self.value * 100.)
        else:
            return "{:13.1f}".format(self.value)

    def format_uncertainty(self):
        """string representation of measurement uncertainty"""
        vs = ""
        if self.stddev is not None:
            if self.is_ratio:
                if self.value != 0.:
                    v = self.stddev * 100.
                else:
                    v = 0.
                vs += "{:.1f}".format(v)
            else:
                vs += "{:6,.1f}".format(self.stddev)
        return vs

    def format_mux(self):
        vs = ""
        if self.multiplex and self.multiplex == self.multiplex:
            vs = "[{:4.1f}%]".format(self.multiplex)
        return vs

    @staticmethod
    def _merge_mux(lhs, rhs):
        return min(lhs.multiplex, rhs.multiplex)

    def update(self, other):
        """merge data from other event into this"""
        assert isinstance(other, UVal), "wrong type"
        # --
        # calc weighted average
        n = self.samples + other.samples
        res = (1. / n) * (self.samples * self + other.samples * other)
        # apply 'res' to this
        self.samples = n
        self.value = res.value
        self.stddev = res.stddev
        self.multiplex = UVal._merge_mux(self, other)

    ######################
    # operators
    ######################

    def ensure_uval(binop): # type: ignore
        """decorator to ensure binary operators are both UVals"""
        def wrapper(self, v): # type: ignore
            if isinstance(v, UVal):
                return binop(self, v) # type: ignore
            elif isinstance(v, (float, int)):
                return binop(self, UVal(TEMPVAL, value=v, stddev=0)) # type: ignore
            else:
                return NotImplemented
        return wrapper

    @ensure_uval # type: ignore
    def __sub__(self, other):
        return UVal._calc(operator.sub, self, other)

    @ensure_uval # type: ignore
    def __add__(self, other):
        return UVal._calc(operator.add, self, other)

    @ensure_uval # type: ignore
    def __mul__(self, other):
        return UVal._calc(operator.mul, self, other)

    @ensure_uval # type: ignore
    def __div__(self, other):
        return UVal._calc(operator.div, self, other) # type: ignore

    @ensure_uval # type: ignore
    def __truediv__(self, other):
        return UVal._calc(operator.truediv, self, other)

    @ensure_uval # type: ignore
    def __lt__(self, other):
        return self.value < other.value

    @ensure_uval # type: ignore
    def __le__(self, other):
        return self.value <= other.value

    @ensure_uval # type: ignore
    def __eq__(self, other):
        return self.value == other.value

    @ensure_uval # type: ignore
    def __ne__(self, other):
        return not self.__eq__(other)

    @ensure_uval # type: ignore
    def __ge__(self, other):
        return self.value >= other.value

    @ensure_uval # type: ignore
    def __gt__(self, other):
        return self.value > other.value

    @ensure_uval # type: ignore
    def __rsub__(self, other):
        """other - self"""
        return UVal._calc(operator.sub, other, self)

    @ensure_uval # type: ignore
    def __radd__(self, other):
        """other + self"""
        return UVal._calc(operator.add, other, self)

    @ensure_uval # type: ignore
    def __rmul__(self, other):
        """other * self"""
        return UVal._calc(operator.mul, other, self)

    @ensure_uval # type: ignore
    def __rdiv__(self, other):
        """other / self"""
        return UVal._calc(operator.div, other, self) # type: ignore

    @ensure_uval # type: ignore
    def __rtruediv__(self, other):
        """other / self"""
        return UVal._calc(operator.truediv, other, self)

    def __nonzero__(self): # python 2
        return self.value != 0.0

    def __bool__(self): # python 3
        return self.value != 0.0

    #########################
    # uncertainty propagator
    #########################

    @staticmethod
    def _calc(op, lhs, rhs, cov=0.):
        """Compute the result of 'lhs [op] rhs' and propagate standard deviations"""
        A = lhs.value
        B = rhs.value
        a = lhs.stddev
        b = rhs.stddev
        # new value
        f = op(float(A), B)
        if isinstance(f, float) and f.is_integer():
            f = int(f)
        # uncertainty
        if op in (operator.mul, operator.truediv, div_op):
            sgn = 1 if op == operator.mul else -1
            if A != 0 and B != 0:
                u = abs(f) * math.sqrt(pow(float(a)/A, 2) + pow(float(b)/B, 2) + sgn*2.*cov/(A*B))
            elif op == operator.mul:
                u = 0.
            elif op == div_op or op == operator.truediv:
                u = 0.
                if A != 0:
                    warn("Error prop failed because of DIV/0: {} {} {}".format(lhs, op, rhs))

        elif op in (operator.add, operator.sub):
            sgn = 1 if op == operator.add else -1
            u = math.sqrt(pow(a, 2) + pow(b, 2) + sgn*2.*cov)

        else:
            u = None
            log.error("Unsupported operation for uncertainty propagator in {} {} {}".format(lhs, op, rhs))
        # --
        ret = UVal(TEMPVAL, value=f, stddev=u, mux=UVal._merge_mux(lhs, rhs), computed=True)
        log.debug("{} {} {} => {}", lhs, op, rhs, ret)
        return ret



# Copyright (c) 2012-2020, Intel Corporation
# Author: Andi Kleen
#
# This program is free software; you can redistribute it and/or modify it
# under the terms and conditions of the GNU General Public License,
# version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# toplev CPU detection

#
# Environment variables to override (mainly for testing):
# FORCECPU=cpu     Force CPU type (e.g. skl). Also --force-cpu
# FORCECOUNTERS=n  Force number of generic counters
# FORCEHT=0/1      Force SMT mode
# HYPERVISOR=0/1   Force hypervisor mode (also --force-hypervisor)
# CPUINFO=file     Read cpu information from file instead of /proc/cpuinfo. Also --force-cpuinfo
# REDUCED_COUNTERS=n Reduce counters by n
#

from collections import defaultdict, Counter
import os
import re
import glob
import sys
if sys.version_info.major == 3:
    import typing # noqa
    from typing import Set, List, Dict, Tuple # noqa

modelid_map = {
    (0x8e, ): "KBLR",
    (0x9e, ): "CFL",
}

cpus_8gpc = set(["icl", "tgl", "icx", "spr", "sprmax", "srf", "lnl", "arl", "gnr"])

def num_offline_cpus():
    cpus = glob.glob("/sys/devices/system/cpu/cpu[0-9]*/online")
    offline = 0
    for fn in cpus:
        with open(fn, "r") as f:
            if int(f.read()) == 0:
                offline += 1
    return offline

def reduced_counters():
    rc = os.getenv("REDUCED_COUNTERS")
    if rc:
        return int(rc)
    val = 1
    fn = "/sys/bus/event_source/devices/cpu/allow_tsx_force_abort"
    if os.path.exists(fn):
        with open(fn, "r") as f:
            val = int(f.read())
    return 1 if val == 0 else 0

class Env(object):
    def __init__(self):
        self.forcecpu = os.getenv("FORCECPU")
        self.forcecounters = os.getenv("FORCECOUNTERS")
        self.forceht = os.getenv("FORCEHT")
        self.hypervisor = os.getenv("HYPERVISOR")
        self.cpuinfo = os.getenv("CPUINFO")

class CPU(object):
    """Detect the CPU."""
    # overrides for easy regression tests
    def force_cpu(self, known_cpus):
        force = self.env.forcecpu
        if not force:
            return False
        self.cpu = ""
        for i in known_cpus:
            if force == i[0]:
                self.cpu = i[0]
                break
        if self.cpu is None:
            print("Unknown FORCECPU ",force)
        return True

    def force_ht(self):
        ht = self.env.forceht
        if ht:
            self.ht = True if int(ht) else False
            return True
        return False

    def __init__(self, known_cpus=(), nocheck=False, env=Env()):
        self.vendor = ""
        self.env = env
        self.model = 0
        self.cpu = ""
        self.realcpu = "simple"
        self.ht = False
        self.counters = {} # type: Dict[str,int]
        self.has_tsx = False
        self.hypervisor = False
        self.force_hypervisor = False
        if self.env.hypervisor:
            self.hypervisor = True
            self.force_hypervisor = True
        self.freq = 0.0
        self.threads = 0
        forced_cpu = self.force_cpu(known_cpus)
        forced_ht = self.force_ht()
        cores = Counter() # type: typing.Counter[Tuple[int,int]]
        sockets = Counter() # type: typing.Counter[int]
        self.coreids = defaultdict(list)
        self.cputocore = {}
        self.cputothread = {}
        self.sockettocpus = defaultdict(list)
        self.cputosocket = {}
        self.allcpus = []
        self.step = 0
        self.name = ""
        cpuinfo = self.env.cpuinfo
        if cpuinfo is None:
            cpuinfo = "/proc/cpuinfo"
        with open(cpuinfo, "r") as f:
            seen = set()
            for l in f:
                n = l.split()
                if len(n) < 3:
                    continue
                if n[0] == 'processor':
                    seen.add("processor")
                    cpunum = int(n[2])
                    self.allcpus.append(cpunum)
                elif (n[0], n[2]) == ("vendor_id", "GenuineIntel"):
                    self.vendor = n[2]
                    seen.add("vendor_id")
                elif (len(n) > 3 and
                        (n[0], n[1], n[3]) == ("cpu", "family", "6")):
                    seen.add("cpu family")
                elif (n[0], n[1]) == ("model", ":"):
                    seen.add("model")
                    self.model = int(n[2])
                elif (n[0], n[1]) == ("model", "name"):
                    seen.add("model name")
                    m = re.search(r"@ (\d+\.\d+)GHz", l)
                    if m:
                        self.freq = float(m.group(1))
                    self.name = " ".join(n[3:])
                elif (n[0], n[1]) == ("physical", "id"):
                    physid = int(n[3])
                    sockets[physid] += 1
                    self.sockettocpus[physid].append(cpunum)
                    self.cputosocket[cpunum] = physid
                elif (n[0], n[1]) == ("core", "id"):
                    coreid = int(n[3])
                    key = (physid, coreid,)
                    cores[key] += 1
                    self.threads = max(self.threads, cores[key])
                    if self.threads > 1 and not forced_ht:
                        self.ht = True
                    self.coreids[key].append(cpunum)
                    self.cputocore[cpunum] = key
                    self.cputothread[cpunum] = self.coreids[key].index(cpunum)
                elif n[0] == "flags":
                    seen.add("flags")
                    self.has_tsx = "rtm" in n
                    if "hypervisor" in n:
                        self.hypervisor = True
                elif n[0] == "stepping":
                    seen.add("stepping")
                    self.step = int(n[2])
        if len(seen) >= 7:
            for i in known_cpus:
                if self.model in i[1] or (self.model, self.step) in i[1]:
                    self.realcpu = i[0]
                    if not forced_cpu:
                        self.cpu = i[0]
                    break
        self.force_counters()
        self.limit4_counters = { "cpu": "none" }
        self.standard_counters = { "cpu": tuple(("0,1,2,3",)) }
        if self.cpu.startswith("lnl"):
            newcounters = {
                "cpu_core": 10,
                "cpu": 10,
                "cpu_atom": 8,
            }
            self.standard_counters = {
                "cpu_core": ("0,1,2,3,4,5,6,7,8,9", ),
                "cpu": ("0,1,2,3,4,5,6,7,8,9", "0,1,2,3,4,5,6,7" ),
                "cpu_atom": ("0,1,2,3,4,5,6,7", )
            }
            self.limit4_counters = { "cpu_core": "0,1,2,3", "cpu_atom": "none",
                                     "cpu": "0,1,2,3" }
        elif self.cpu.startswith("adl") or self.cpu.startswith("mtl"):
            atom_counters = 6 if self.cpu.startswith("adl") else 8
            newcounters = {
                "cpu_core": 8,
                "cpu": 4,
                "cpu_atom": atom_counters,
            }
            self.standard_counters = {
                "cpu_core": ("0,1,2,3,4,5,6,7", "0,1,2,3", ),
                "cpu": ("0,1,2,3,4,5,6,7", "0,1,2,3", "0,1,2,3,4,5" ),
                "cpu_atom": ("0,1,2,3,4,5", ) if atom_counters == 6 else ("0,1,2,3,4,5,6,7,", )
            }
            self.limit4_counters = { "cpu_core": "0,1,2,3", "cpu_atom": "none",
                                     "cpu": "0,1,2,3" }
        elif self.cpu == "slm":
            newcounters = { "cpu": 2 }
            self.standard_counters = { "cpu": ("0,1",) }
        # when running in a hypervisor always assume worst case HT in on
        # also when CPUs are offline assume SMT is on
        elif self.ht or self.hypervisor or (num_offline_cpus() > 0 and not nocheck) or self.cpu in cpus_8gpc:
            if self.cpu in cpus_8gpc or (self.cpu == "simple" and self.realcpu in cpus_8gpc):
                newcounters = {"cpu": 8 }
                self.standard_counters = { "cpu": ("0,1,2,3,4,5,6,7", "0,1,2,3", ) }
                self.limit4_counters = { "cpu": "0,1,2,3" }
            else:
                newcounters = { "cpu": 4 }
        elif self.cpu == "ehl":
            newcounters = { "cpu": 4 }
        else:
            newcounters = { "cpu": 8 }
        if not self.counters:
            self.counters = newcounters
        if not nocheck and not self.env.forcecounters:
            for j in ("cpu", "cpu_core", "cpu_atom"):
                if j in self.counters:
                    self.counters[j] -= reduced_counters()
        if self.cpu in cpus_8gpc:
            self.standard_counters = { "cpu": ("0,1,2,3,4,5,6,7", "0,1,2,3",
                                               "0,1,2,3,4,5,6,7", "0,1,2,3,4,5,6,7,8,9") }
            self.limit4_counters = { "cpu": "0,1,2,3" }

        if self.cpu == "spr" and "Max" in self.name:
            self.cpu = "sprmax"

        try:
            self.pmu_name = open("/sys/bus/event_source/devices/cpu/caps/pmu_name").read().strip()
        except IOError:
            self.pmu_name = ""

        self.sockets = len(sockets.keys())
        self.modelid = None
        mid = (self.model,)
        self.true_name = self.cpu
        if mid in modelid_map:
            self.modelid = modelid_map[mid]
            self.true_name = self.modelid.lower()
        # XXX match steppings here too

    def force_counters(self):
        cnt = self.env.forcecounters
        if cnt:
            cntn = int(cnt)
            if self.realcpu in ("adl", "mtl", "lnl"):
                self.counters = {
                    "cpu_core": cntn,
                    "cpu_atom": cntn,
                    "cpu": cntn }
            else:
                self.counters = { "cpu": cntn }




# dummy arithmetic type without any errors, for collecting
# the events from the model. Otherwise divisions by zero cause
# early exits

class DummyArith(object):
    def __add__(self, o):
        return self
    __sub__ = __add__
    __mul__ = __add__
    __div__ = __add__
    __truediv__ = __add__
    __rsub__ = __add__
    __radd__ = __add__
    __rmul__ = __add__
    __rdiv__ = __add__
    __rtruediv__ = __add__
    def __lt__(self, o):
        return True
    __eq__ = __lt__
    __ne__ = __lt__
    __gt__ = __lt__
    __ge__ = __lt__
    __or__ = __add__
    __and__ = __add__
    __min__ = __add__
    __max__ = __add__



#!/usr/bin/env python
# resolve kernel symbols through kallsyms (when no vmlinux is available)
#
# Copyright (c) 2014, Intel Corporation
# Author: Andi Kleen
#
# This program is free software; you can redistribute it and/or modify it
# under the terms and conditions of the GNU General Public License,
# version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.

import util

kernel = []

def parse_kernel():
    with open("/proc/kallsyms", 'r') as f:
        for l in f:
            n = l.split()
            addr = int(n[0], 16)
            kernel.append((addr, n[2]))

def resolve_kernel(ip):
    if not kernel:
        parse_kernel()
    n = util.find_le(kernel, ip)
    if n:
        return n[1], ip - n[0]
    return None



#!/usr/bin/env python
# sanity check an event
# percent between 0 and 100%
# nothing negative
from __future__ import print_function
import sys
import os
# thanks python 3.13!
try:
    from shlex import quote as cmd_quote
except ImportError:
    from pipes import quote as cmd_quote

logfile = "slog.%d" % (os.getpid())

s = "./ucevent.py -x, -o " + logfile + " " + " ".join(map(cmd_quote, sys.argv[1:]))
w = os.getenv("WRAP")
if w:
    s = w + " " + s
print(s)
r = os.system(s)
if r != 0:
    print("ucevent failed", r)
    sys.exit(1)

f = open(logfile, "r")
fields = f.readline().strip().split(",")
for l in f:
    vals = l.strip().split(",")
    for v, h in zip(vals, fields):
        if fields == "timestamp":
            continue
        try:
            num = float(v)
        except ValueError:
            print(h,v)
            continue
        if num < 0:
            print(h,"negative value",v)
        if h.find("_PCT") >= 0 or h.find("PCT_") >= 0:
            if num < 0 or num > 1.01:
                print(h,"percent out of bound", v)

os.remove(logfile)



# distinguish the bewildering variety of perf/toplev CSV formats
from __future__ import print_function
import sys
import re
from collections import namedtuple

def is_val(n):
    return re.match(r'-?[0-9.]+%?|<.*>', n) is not None

def is_cpu(n):
    return re.match(r'(CPU)|(S\d+(-C\d+)?)|C\d+|all', n) is not None

def is_socket(n):
    return re.match(r'S\d+', n) is not None

def is_event(n):
    return re.match(r'[a-zA-Z.-]+', n) is not None

def is_number(n):
    return re.match(r'\s*[0-9]+', n) is not None

def is_ts(n):
    return re.match(r'\s*[0-9.]+', n) is not None or n == "SUMMARY"

def is_unit(n):
    return re.match(r'(% )?[a-zA-Z]*( <)?', n) is not None

def is_running(n):
    return is_number(n)

def is_enabled(n):
    return is_number(n)

formats = (
# 0.100997872;CPU0;4612809;;inst_retired_any_0;3491526;2.88     new perf
        (is_ts, is_cpu, is_val, is_unit, is_event, is_enabled, is_running),
# 1.354075473,0,cpu-migrations                            old perf w/o cpu
        (is_ts, is_val, is_event),
# 1.354075473,CPU0,0,cpu-migrations                          old perf w/ cpu
        (is_ts, is_cpu, is_val, is_event),
# 0.799553738,137765150,,branches                              new perf with unit
        (is_ts, is_val, is_unit, is_event),
# 0.799553738,CPU1,137765150,,branches                  new perf with unit and cpu
        (is_ts, is_cpu, is_val, is_unit, is_event),
# 0.100879059,402.603109,,task-clock,402596410,100.00    new perf with unit without cpu and stats
        (is_ts, is_val, is_unit, is_event, is_running, is_enabled),
# 1.001131873,S0,Backend_Bound.Memory_Bound,13.3,% Slots <,,,0.0,3.0,,
# 0.200584389,0,FrontendBound.Branch Resteers,15.87%,above,"",  toplev w/ cpu
        (is_ts, is_cpu, is_event, is_val, is_unit),
# 1.001365014,CPU2,1819888,,instructions,93286388,100.00      new perf w/ unit w/ cpu and stats
        (is_ts, is_cpu, is_val, is_unit, is_event, is_running, is_enabled),
# 0.609113353,S0,4,405.454531,,task-clock,405454468,100.00      perf --per-socket with cores
        (is_ts, is_socket, is_number, is_val, is_unit, is_event, is_running, is_enabled),
# 0.806231582,S0,4,812751,,instructions                  older perf --per-socket w/ cores w/o stats
        (is_ts, is_socket, is_number, is_val, is_unit, is_event),
# 0.936482669,C1-T0,Frontend_Bound.Frontend_Latency.ITLB_Misses,0.39,%below,,itlb_misses.walk_completed,,
# 0.301553743,C1,Retiring,31.81,%,,,,
# 0.200584389,FrontendBound.Branch Resteers,15.87%,above,"",    toplev single thread
        (is_ts, is_event, is_val),
)

fmtmaps = {
    is_ts: 0,
    is_cpu: 1,
    is_event: 2,
    is_val: 3,
    is_enabled: 4,
    is_running: 5,
    is_unit: 6
}

Row = namedtuple('Row', ['ts', 'cpu', 'ev', 'val', 'enabled', 'running', 'unit'])

def check_format(fmt, row):
    if all([x(n.strip()) for (x, n) in zip(fmt, row)]):
        vals = [None] * 7
        for i, j in enumerate(fmt):
            if j in fmtmaps:
                vals[fmtmaps[j]] = row[i]
        r = Row._make(vals)
        return r
    return False

fmt_cache = formats[0]

def parse_csv_row(row, error_exit=False):
    if len(row) == 0:
        return None
    global fmt_cache
    r = check_format(fmt_cache, row)
    if r:
        return r
    for fmt in formats:
        r = check_format(fmt, row)
        if r:
            fmt_cache = fmt
            return r
    if row[0].startswith("#"):    # comment
        return None
    if ".csv" in row[0]:          # fake-perf output
        return None
    if "Timestamp" in row[0]:
        return None
    print("PARSE-ERROR", row, file=sys.stderr)
    if error_exit:
        sys.exit(1)
    return None

if __name__ == '__main__':
    def check(l, fields):
        n = l.split(",")
        r = parse_csv_row(n)
        assert r is not None
        rd = r._asdict()
        for a, v in fields.items():
            assert rd[a] == n[v]

    check('1.001131873,S0,Backend_Bound.Memory_Bound,13.3,% Slots <,,,0.0,3.0,,', {
            "ts": 0,
            "cpu": 1,
            "ev": 2,
            "val": 3,
            "unit": 4 })



#
# Simple 5 event top level model
#

print_error = lambda msg: False
version = "1.0"
check_event = lambda ev: False

# Constants

PipelineWidth = 4

uops_retired_slots = "UOPS_RETIRED.RETIRE_SLOTS"

def CLKS(EV):
    return EV("CPU_CLK_UNHALTED.THREAD", 1)

def SLOTS(EV):
    return PipelineWidth * CLKS(EV)

# Instructions Per Cycle
def IPC(EV, level):
    return EV("INST_RETIRED.ANY", level) / CLKS(EV)

# Uops Per Instruction
def UPI(EV, level):
    return EV(uops_retired_slots, level) / EV("INST_RETIRED.ANY", level)

# Average Frequency Utilization relative nominal frequency
def TurboUtilization(EV, level):
    return CLKS(EV) / EV("CPU_CLK_UNHALTED.REF_TSC", level)

class FrontendBound:
    name = "Frontend Bound"
    domain = "Slots"
    desc = """
This category reflects slots where the Frontend of the processor undersupplies
its Backend."""
    level = 1
    def compute(self, EV):
        try:
            self.val = EV("IDQ_UOPS_NOT_DELIVERED.CORE", 1) / SLOTS(EV)
            self.thresh = self.val > 0.2
        except ZeroDivisionError:
            self.val = 0
            self.thresh = False
        return self.val

class BadSpeculation:
    name = "Bad Speculation"
    domain = "Slots"
    desc = """
This category reflects slots wasted due to incorrect speculations, which
include slots used to allocate uops that do not eventually get retired and
slots for which allocation was blocked due to recovery from earlier incorrect
speculation. For example, wasted work due to miss-predicted branches is
categorized under the Bad Speculation category"""
    level = 1
    def compute(self, EV):
        try:
            self.val = ( EV("UOPS_ISSUED.ANY", 1) - EV(uops_retired_slots, 1) + PipelineWidth * EV("INT_MISC.RECOVERY_CYCLES", 1) ) / SLOTS(EV)
            self.thresh = self.val > 0.1
        except ZeroDivisionError:
            self.val = 0
            self.thresh = False
        return self.val

class BackendBound:
    name = "Backend Bound"
    domain = "Slots"
    desc = """
This category reflects slots where no uops are being delivered due to a lack
of required resources for accepting more uops in the Backend of the pipeline."""
    level = 1
    def compute(self, EV):
        try:
            self.val = 1 - ( self.FrontendBound.compute(EV) + self.BadSpeculation.compute(EV) + self.Retiring.compute(EV) )
            self.thresh = self.val > 0.2
        except ZeroDivisionError:
            self.val = 0
            self.thresh = False
        return self.val

class Retiring:
    name = "Retiring"
    domain = "Slots"
    desc = """
This category reflects slots utilized by good uops i.e. allocated uops that
eventually get retired."""
    level = 1
    def compute(self, EV):
        try:
            self.val = EV(uops_retired_slots, 1) / SLOTS(EV)
            self.thresh = self.val > 0.7
        except ZeroDivisionError:
            self.val = 0
            self.thresh = False
        return self.val

class Metric_IPC:
    name = "IPC"
    desc = """
Instructions Per Cycle"""
    errcount = 0

    def compute(self, EV):
        try:
            self.val = IPC(EV, 0)
        except ZeroDivisionError:
            print_error("IPC zero division")
            self.errcount += 1
            self.val = 0

class Metric_UPI:
    name = "UPI"
    desc = """
Uops Per Instruction"""
    errcount = 0

    def compute(self, EV):
        try:
            self.val = UPI(EV, 0)
        except ZeroDivisionError:
            print_error("UPI zero division")
            self.errcount += 1
            self.val = 0

class Metric_TurboUtilization:
    name = "TurboUtilization"
    desc = """
Average Frequency Utilization relative nominal frequency"""
    errcount = 0

    def compute(self, EV):
        try:
            self.val = TurboUtilization(EV, 0)
        except ZeroDivisionError:
            print_error("TurboUtilization zero division")
            self.errcount += 1
            self.val = 0

class Setup:
    def __init__(self, r):
        if check_event("UOPS_RETIRED.SLOTS"):
            global uops_retired_slots
            uops_retired_slots = "UOPS_RETIRED.SLOTS"

        o = dict()
        n = FrontendBound() ; r.run(n)
        o["FrontendBound"] = n
        n = BadSpeculation() ; r.run(n)
        o["BadSpeculation"] = n
        n = BackendBound() ; r.run(n)
        o["BackendBound"] = n
        n = Retiring() ; r.run(n)
        o["Retiring"] = n

        o["FrontendBound"].parent = None
        o["BadSpeculation"].parent = None
        o["BackendBound"].parent = None
        o["Retiring"].parent = None
        o["BackendBound"].FrontendBound = o["FrontendBound"]
        o["BackendBound"].BadSpeculation = o["BadSpeculation"]
        o["BackendBound"].Retiring = o["Retiring"]

        o["FrontendBound"].sibling = None
        o["BadSpeculation"].sibling = None
        o["BackendBound"].sibling = None
        o["Retiring"].sibling = None

        o["FrontendBound"].sample = []
        o["BadSpeculation"].sample = []
        o["BackendBound"].sample = []
        o["Retiring"].sample = []

        # user visible metrics

        n = Metric_IPC() ; r.metric(n)
        n = Metric_UPI() ; r.metric(n)
        n = Metric_TurboUtilization() ; r.metric(n)



#!/usr/bin/env python3
# -*- coding: utf-8
# generate return latencies to tune toplev model
# Copyright (c) 2023, Intel Corporation
# Author: Andi Kleen
#
# This program is free software; you can redistribute it and/or modify it
# under the terms and conditions of the GNU General Public License,
# version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
from __future__ import print_function
import subprocess as subp
import os
import sys
import ocperf
import json
import argparse
import random
import glob
import re
from dummyarith import DummyArith
from collections import Counter, defaultdict
from copy import copy
import csv

def cmd(args, l):
    if args.verbose:
        print(" ".join(l))
    return l

class SamplePerfRun(object):
    """Run perf record to collect Timed PEBS information and generate averages for event weights."""
    def __init__(self, args):
        self.pi = self.pr = self.ps = None
        self.args = args

    def execute(self, perf, evsamples, origevents, pargs):
        self.perf = perf
        self.rmap = { x: y for x, y in zip(evsamples, origevents) }
        pr = subp.Popen(cmd(self.args, [perf, "record",
                                   "-W",
                                   "-B",
                                   "-c", "%d" % self.args.interval,
                                   "-e", ",".join(evsamples),
                                   "-o", "-"] + pargs),
                                   stdout=subp.PIPE)
        pi = subp.Popen(cmd(self.args, [perf, "inject"]),
                                 stdin=pr.stdout,
                                 stdout=subp.PIPE)
        ps = subp.Popen(cmd(self.args, [perf, "script",
                                   "-i", "-",
                                   "--ns",
                                   "-F", "time,retire_lat,event"]),
                                   stdin=pi.stdout,
                                   stdout=subp.PIPE)
        pi.stdout.close()
        pr.stdout.close()
        self.pr, self.pi, self.ps = pr, pi, ps

    def handle_samples(self):
        for l in self.ps.stdout:
            l = l.decode()
            n = l.split()
            ts, event, weight = float(n[0].replace(":","")), n[1].replace(":",""), int(n[2])
            if self.args.csvplot:
                self.args.csvplot.writerow([ts, self.rmap[event] if event in self.rmap else event, weight])
            if self.args.fake:
                yield event, random.randint(0, 100)
            else:
                yield event, weight
        self.pr.wait()
        self.pi.wait()
        self.ps.wait()

NUM_SPARKS = 8
SPARK_SHIFT = 8

def spark_f(l):
    min_ = min(l)
    f = int(max(((max(l) - min_) << SPARK_SHIFT) / (NUM_SPARKS - 1), 1))
    return f, min_

# this samples unlike normal spark lines
def gen_spark_buckets(l):
    if len(l) == 0:
        return [], 0, 1
    f, min_ = spark_f(l)
    return [((x - min_) << SPARK_SHIFT) / f for x in random.sample(l, min(NUM_SPARKS, len(l)))], min_, f

def lookup(s, i, fb):
    if 0 <= i <= len(s) - 1:
        return s[i]
    return fb

def gen_spark(buckets, min_, f):
    return "".join([lookup("▁▂▃▄▅▆▇█", int((int((x - min_)) << SPARK_SHIFT) / f), " ") for x in buckets])

def gen_stat(samples):
    # {
    #     "COUNT": 5358917,
    #     "MIN": 0,
    #     "MAX": 65535,
    #     "MEAN": 3.23,
    #     "MEDIAN": 0,
    #     "NZ_MEDIAN": 1,
    #     "MODE": 0,
    #     "MODE_COUNT": 3631698,
    #     "NZ_MODE": 1,
    #     "NZ_MODE_COUNT": 1213029,
    #     "BUCKETS": 2344
    #   },
    nz = [x for x in samples if x != 0.0]
    buckets = Counter(samples)
    nz_buckets = copy(buckets)
    if 0 in nz_buckets:
        del nz_buckets[0]
    spark, min_, f = gen_spark_buckets(samples)
    spark_nz, min_nz, f_nz = gen_spark_buckets(nz)
    return {
        "COUNT": len(samples),
        "MIN": min(samples),
        "MAX": max(samples),
        "MEAN": round(float(sum(samples)) / len(samples), 2),
        "MEDIAN": sorted(samples)[len(samples)//2] if len(samples) > 0 else 0.0,
        "NZ_MEDIAN": sorted(nz)[len(nz)//2] if len(nz) > 0 else 0.0,
        "MODE": buckets.most_common(1)[0][0] if len(buckets) > 0 else 0.0,
        "MODE_COUNT": buckets.most_common(1)[0][1] if len(buckets) > 0 else 0,
        "NZ_MODE": nz_buckets.most_common(1)[0][0] if len(nz_buckets) > 0 else 0,
        "NZ_MODE_COUNT": nz_buckets.most_common(1)[0][1] if len(nz_buckets) > 0 else 0,
        "BUCKETS": len(buckets),
        "F": f,
        "F_NZ": f_nz,
        "MIN_NZ": min_nz,
        "SPARK_BUCKETS": ",".join(["%d" % x for x in spark]),
        "SPARK_BUCKETS_NZ": ",".join(["%d" % x for x in spark_nz]),
    }

def human_output(data):
    d = data["Data"]
    for ev in sorted(d.keys()):
        print("%s: " % ev, end="")
        for m in sorted(d[ev].keys()):
            if m.startswith("SPARK"):
                if d[ev][m] == "":
                    continue
                l = [(int(x) if x.isdecimal() else 0) for x in d[ev][m].split(",")]
                s = gen_spark(l, d[ev]["MIN"], d[ev]["F_NZ" if m.endswith("_NZ") else "F"])
                print("%s %s " % (m.lower(), s), end="")
            else:
                print("%s %s " % (m.lower(), d[ev][m]), end="")
        print()

def get_model_number():
    with open("/proc/cpuinfo") as f:
        for l in f:
            n = l.split()
            if len(n) >= 3 and n[0] == "model" and n[1] == ":":
                return int(n[2])
    return 0

def find_model(args):
    if not args.cpu:
        cpu = "?"
        cl = glob.glob("/sys/bus/event_source/devices/cpu*/caps/pmu_name")
        for fn in cl:
            cpu = open(fn).read().strip()
            if cpu == "meteorlake_hybrid":
                args.cpu = "mtl"
            elif cpu == "sapphire_rapids":
                model = get_model_number()
                if model == 0xad or model == 0xae:
                    args.cpu = "gnr"
                else:
                    sys.exit("Unsupported CPU %d" % model)
            elif cpu == "granite_rapids":
                args.cpu = "gnr"
            elif cpu == "lunarlake_hybrid":
                args.cpu = "lnl"
    if not args.cpu:
        sys.exit("Unsupported CPU %s" % cpu)
    if args.cpu == "mtl":
        import mtl_rwc_ratios
        return mtl_rwc_ratios
    elif args.cpu == "gnr":
        import gnr_server_ratios
        return gnr_server_ratios
    elif args.cpu == "lnl":
        import lnl_lnc_ratios
        return lnl_lnc_ratios
    sys.exit("Unknown cpu %s" % args.cpu)

def gen_events(args):
    model = find_model(args)

    nodes = []
    class Runner(object):
        def metric(self, n):
            nodes.append(n)
            n.thresh = True
        def run(self, n):
            nodes.append(n)
            n.thresh = True

    model.Setup(Runner())

    events = set()
    def collect(name, level):
        if level == 999:
            events.add(name + ":pp")
        return DummyArith()

    for n in nodes:
        n.compute(collect)

    return sorted(events)

def getevent(emap, e):
    ev = emap.getevent(e)
    if ev is None:
        print(e, "not found")
        return "dummy"
    return ev.output()

def init_args():
    ap = argparse.ArgumentParser('Generate JSON of retirement latencies to tune toplev')
    ap.add_argument('--output', '-o', type=argparse.FileType('w'), default=sys.stdout,
                    help="")
    ap.add_argument('--verbose', '-v', action='store_true', help='be verbose')
    ap.add_argument('--fake', action='store_true', help=argparse.SUPPRESS)
    ap.add_argument('--interval', '-i', type=int, default=1003, help="Interval for sampling")
    ap.add_argument('--pmu', '-p', nargs='*', default=["cpu", "cpu_core"], help="for which PMUs to collect")
    ap.add_argument('--quiet', '-q', action='store_true')
    ap.add_argument('--csv', '-c', type=argparse.FileType('w'), help="Generate CSV file with pushout latencies", default=None)
    ap.add_argument('--cpu', help="Set CPU type (gnr, mtl, lnl)")
    args, rest = ap.parse_known_args()
    if args.csv:
        args.csvplot = csv.writer(args.csv)
    else:
        args.csvplot = None
    return args, rest

def main():
    args, rest = init_args()
    events = gen_events(args)
    pmus = ocperf.find_pmus()
    emap = ocperf.find_emap(pmu=[x for x in pmus if x != "cpu_atom"][0])
    if emap is None:
        sys.exit("Cannot find json event list")
    pevents = [getevent(emap, e) for e in events]
    if args.verbose:
        print(events)
    s = SamplePerfRun(args)
    perf = os.getenv("PERF")
    if perf is None:
        perf = "perf"
    try:
        s.execute(perf, pevents, events, rest)
        samples = defaultdict(list)
        for ev, weight in s.handle_samples():
            samples[ev].append(weight)
    except KeyboardInterrupt:
        pass
    data = { "Data": { re.sub(r"[:/][uU]?", "", ev.upper()).replace("CPU_CORE","").replace("RETIRED_", "RETIRED."): gen_stat(s)
                       for ev, s in samples.items()
                       if "/" not in ev or any([x in ev for x in args.pmu]) } }
    json.dump(data, args.output, indent=2, sort_keys=True)
    if not args.quiet:
        human_output(data)

if __name__ == '__main__':
    main()



EVENTSEL_EVENT = 0x00ff
EVENTSEL_UMASK = 0xff00
EVENTSEL_UMASK2 = 0xff00000000
EVENTSEL_EDGE  = 1<<18
EVENTSEL_PC    = 1<<19
EVENTSEL_ANY   = 1<<21
EVENTSEL_INV   = 1<<23
EVENTSEL_EQ    = 1<<36
EVENTSEL_INTX  = 1<<32
EVENTSEL_INTX_CP = 1<<33
EVENTSEL_CMASK = 0xff000000

EVMASK = (EVENTSEL_EVENT | EVENTSEL_UMASK | EVENTSEL_EDGE | EVENTSEL_PC | EVENTSEL_ANY |
          EVENTSEL_INV | EVENTSEL_CMASK | EVENTSEL_UMASK2 | EVENTSEL_EQ | EVENTSEL_INTX |
          EVENTSEL_INTX_CP)

EVENTSEL_ENABLE = 1<<22

MSR_EVNTSEL = 0x186
MSR_IA32_FIXED_CTR_CTRL = 0x38d
MSR_PEBS_ENABLE = 0x3f1
MSR_PERFCTR = 0xc1
MSR_PMC = 0x4c1
MSR_FIXED_CTR = 0x309
MSR_FIXED_CTR_CTL = 0x38d
MSR_GLOBAL_STATUS = 0x38e
MSR_GLOBAL_CTRL = 0x38f
MSR_GLOBAL_OVF_CTRL = 0x390

extra_flags = (
        (EVENTSEL_EDGE, "edge"),
        (EVENTSEL_PC, "pc"),
        (EVENTSEL_ANY, "any"),
        (EVENTSEL_INV, "inv"),
        (EVENTSEL_CMASK, "cmask"))



# Copyright (c) 2012-2020, Intel Corporation
# Author: Andi Kleen
#
# This program is free software; you can redistribute it and/or modify it
# under the terms and conditions of the GNU General Public License,
# version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# Output toplev results in various formats

from __future__ import print_function
import locale
import csv
import re
import sys
import json
import os
from math import isnan
from collections import defaultdict, Counter, OrderedDict
from tl_uval import UVal, combine_uval
from tl_io import flex_open_w
if sys.version_info.major == 3:
    import typing # noqa
    from typing import DefaultDict, Dict, Any # noqa

def output_name(name, typ):
    if typ:
        if "." in name:
            name = re.sub(r'(.*)\.', r'\1-%s.' % typ, name)
        else:
            name += "-" + typ
    return name

def open_logfile(name, typ):
    if name is None or name == "":
        return sys.stderr
    if 'write' in name.__class__.__dict__:
        return name
    name = output_name(name, typ)
    try:
        return flex_open_w(name)
    except IOError:
        sys.exit("Cannot open logfile %s" % name)

def open_all_logfiles(args, logfile):
    if args.split_output and args.per_thread + args.per_core + args.per_socket + args.global_ > 0:
        logfiles = dict()
        if args.per_thread:
            logfiles['thread'] = open_logfile(logfile, "thread")
        if args.per_core:
            logfiles['core'] = open_logfile(logfile, "core")
        if args.per_socket:
            logfiles['socket'] = open_logfile(logfile, "socket")
        if args.global_:
            logfiles['global'] = open_logfile(logfile, "global")
        return logfiles, None
    else:
        return None, open_logfile(logfile, None)

BUFS = 1024*1024

def catrmfile(infn, outf, keep):
    with open(infn, "r") as inf:
        while True:
            buf = inf.read(BUFS)
            if len(buf) == 0:
                break
            outf.write(buf)
    outf.flush()
    if not keep:
        os.remove(infn)

def catrmoutput(infn, logf, logfiles, keep):
    if logfiles:
        for j in logfiles.keys():
            catrmfile(output_name(infn, j), logfiles[j], keep)
    else:
        catrmfile(infn, logf, keep)

class Output(object):
    """Abstract base class for Output classes."""
    def __init__(self, logfile, version, cpu, args):
        self.logfiles, self.logf = open_all_logfiles(args, args.output)
        self.printed_descs = set()
        self.hdrlen = 30
        self.version = version
        self.unitlen = 12
        self.belowlen = 0
        self.version = "%s on %s [%s%s]" % (version, cpu.name, cpu.true_name,
                "/" + cpu.pmu_name if cpu.pmu_name else "")
        self.curname = ""
        self.curname_nologf = ""
        self.printedversion = set()
        self.no_header = args.no_csv_header
        self.no_footer = args.no_csv_footer
        self.abbrev = args.abbrev
        self.valcsv = None
        self.last_prefix = ""
        self.args = args

    def flushfiles(self):
        if self.logfiles:
            for j in self.logfiles.values():
                j.flush()
        self.logf.flush()

    # pass all possible hdrs in advance to compute suitable padding
    def set_hdr(self, hdr, area):
        if area:
            hdr = "%-16s %s" % (area, hdr)
        self.hdrlen = max(len(hdr) + 1, self.hdrlen)

    def set_below(self, below):
        if below:
            self.belowlen = 1

    def set_unit(self, unit):
        self.unitlen = max(len(unit), self.unitlen)

    def set_cpus(self, cpus):
        pass

    def item(self, area, name, uval, timestamp, unit, desc, title, sample, bn, below, idle):
        assert isinstance(uval, UVal)
        # --
        if desc in self.printed_descs:
            desc = ""
        else:
            self.printed_descs.add(desc)
        if not area:
            area = ""
        self.show(timestamp, title, area, name, uval, unit, desc, sample, bn, below, idle)

    def ratio(self, area, name, uval, timestamp, unit, desc, title, sample, bn, below, idle):
        uval.is_ratio = True
        self.item(area, name, uval, timestamp, unit, desc, title, sample, bn, below, idle)

    def metric(self, area, name, uval, timestamp, desc, title, unit, idle):
        self.item(area, name, uval, timestamp, unit, desc, title, None, "", "", idle)

    def flush(self):
        pass

    def remark(self, m):
        if not self.logfiles:
            self.logf.write('\n%s:\n' % m)

    def reset(self, name):
        if self.logfiles:
            self.logf = self.logfiles[name]
            self.curname = name
        self.curname_nologf = name

    def show(self, timestamp, title, area, hdr, val, unit, desc, sample, bn, below, idle):
        pass

    def print_version(self):
        if self.no_header or self.args.no_version:
            return
        if self.curname not in self.printedversion:
            if self.logfiles:
                self.logfiles[self.curname].write("# " + self.version + "\n")
            else:
                self.logf.write("# " + self.version + "\n")
            self.printedversion.add(self.curname)

    print_header = print_version

    def print_footer(self):
        pass

    def print_footer_all(self):
        if self.no_footer or self.args.no_version:
            return
        if self.logfiles:
            for f in self.logfiles.values():
                f.write("# %s\n" % self.version)
        else:
            self.logf.write("# " + self.version + "\n")

def fmt_below(below):
    if below:
        return "<"
    return ""

def short_hdr(hdr, last):
    n = os.path.commonprefix((hdr, last))
    if "." not in n:
        return hdr
    n = n[:n.rfind(".")]
    return "..." + hdr[len(n)+1:]

class OutputHuman(Output):
    """Generate human readable single-column output."""
    def __init__(self, logfile, args, version, cpu):
        Output.__init__(self, logfile, version, cpu, args)
        try:
            locale.setlocale(locale.LC_ALL, '')
        except locale.Error:
            pass
        self.args = args
        self.titlelen = 7

    def set_cpus(self, cpus):
        if len(cpus) > 0:
            self.titlelen = max(map(len, cpus)) + 1

    def print_desc(self, desc, sample):
        if self.args.no_desc:
            return
        if desc:
            print("\t" + desc, file=self.logf)
            if sample:
                print("\t" + "Sampling events: ", sample, file=self.logf)

    def print_timestamp(self, timestamp):
        if timestamp:
            if isnan(timestamp):
                self.logf.write("%-11s " % "SUMMARY")
            else:
                self.logf.write("%6.9f " % timestamp)

    def print_line_header(self, area, ohdr):
        if "Info" in area or not self.abbrev:
            hdr = ohdr
        else:
            hdr = short_hdr(ohdr, self.last_prefix)
            self.last_prefix = ohdr
        if area:
            hdr = "%-16s %s" % (area, hdr)
        self.logf.write("%-*s " % (self.hdrlen, hdr))

    # timestamp Timestamp in interval mode
    # title     CPU
    # area      FE/BE ...
    # hdr       Node Name
    # val       Formatted measured value
    # unit      unit
    # desc      Object description
    # sample    Sample Objects (string)
    # vs        Statistics object
    # bn        marker for bottleneck
    # below     True if below
    # idle      Idle marker (ignored for Human)
    # Example:
    # C0    BE      Backend_Bound:                                62.00 %
    def show(self, timestamp, title, area, hdr, val, unit, desc, sample, bn, below, idle):
        self.print_header()
        self.print_timestamp(timestamp)
        write = self.logf.write
        if title:
            write("%-*s" % (self.titlelen, title))
        self.print_line_header(area, hdr)
        vals = "{:<{unitlen}} {:>20} {:<{belowlen}}".format(
                    ("  " if unit and unit[0] != "%" else "") + unit,
                    val.format_value(unit),
                    fmt_below(below),
                    unitlen=self.unitlen + 2,
                    belowlen=self.belowlen)
        if not self.args.no_mux and val.multiplex != 100.0:
            vals += " " + val.format_mux()
        if val.stddev:
            vals += " +- {:>8}".format(val.format_uncertainty())
        if bn:
            vals += bn
        write(vals + "\n")
        self.print_desc(desc, sample)

    def metric(self, area, name, uval, timestamp, desc, title, unit, idle):
        self.item(area, name, uval, timestamp, unit, desc, title, None, "", "", False)

def convert_ts(ts):
    if isnan(ts):
        return "SUMMARY"
    return ts

class OutputColumns(OutputHuman):
    """Human-readable output data in per-cpu columns."""
    def __init__(self, logfile, args, version, cpu):
        OutputHuman.__init__(self, logfile, args, version, cpu)
        self.nodes = {}
        self.timestamp = None
        self.cpunames = set()
        self.printed_header = False

    def set_cpus(self, cpus):
        self.cpunames = cpus

    def show(self, timestamp, title, area, hdr, val, unit, desc, sample, bn, below, idle):
        if self.args.single_thread:
            OutputHuman.show(self, timestamp, title, area, hdr, val, unit, desc, sample, bn, below, idle)
            return
        self.print_header()
        self.timestamp = timestamp
        key = (area, hdr)
        if key not in self.nodes:
            self.nodes[key] = {}
        assert title not in self.nodes[key]
        self.nodes[key][title] = (val, unit, desc, sample, bn, below, idle)

    def flush(self):
        VALCOL_LEN = 16
        write = self.logf.write

        cpunames = sorted(self.cpunames)

        if not self.printed_header:
            if self.timestamp:
                write("%9s" % "")
            self.print_line_header("", "")
            for j in cpunames:
                write("%*s  " % (VALCOL_LEN, j))
            write("\n")
            self.printed_header = True

        for key in sorted(sorted(self.nodes.keys(), key=lambda x: x[1]), key=lambda x: x[0] == ""):
            node = self.nodes[key]
            desc = None
            sample = None
            unit = None
            if self.timestamp:
                self.print_timestamp(self.timestamp)

            self.print_line_header(key[0], key[1])
            vlist = []
            for cpuname in cpunames:
                if cpuname in node:
                    cpu = node[cpuname]
                    uval, unit, desc, sample, bn, below, idle = cpu
                    v = uval.format_value(unit)
                    vlist.append(uval)
                    write("%*s%s " % (VALCOL_LEN, v, "?" if below else "*" if bn else " "))
                else:
                    write("%*s  " % (VALCOL_LEN, ""))
            if unit:
                # XXX should move this to be per entry?
                cval = combine_uval(vlist)
                vs = (" +- " + cval.format_uncertainty() + " " + cval.format_mux()) if cval.stddev else ""
                write(" %-*s%s" % (self.unitlen, ("  " if unit[0] != "%" else "") + unit, vs))
            write("\n")
            self.print_desc(desc, sample)
        self.nodes = {}

    def reset(self, name):
        Output.reset(self, name)
        self.printed_header = False

class OutputColumnsCSV(OutputColumns):
    """Columns output in CSV mode."""

    def __init__(self, logfile, sep, args, version, cpu):
        OutputColumns.__init__(self, logfile, args, version, cpu)
        self.writer = {}
        if self.logfiles:
            for n, f in self.logfiles.items():
                self.writer[n] = csv.writer(f, delimiter=sep, lineterminator='\n')
        else:
            self.writer[''] = csv.writer(self.logf, delimiter=sep, lineterminator='\n')
        self.printed_header = False

    # XXX implement bn and idle
    def show(self, timestamp, title, area, hdr, val, unit, desc, sample, bn, below, idle):
        self.print_header()
        self.timestamp = timestamp
        key = (area, hdr)
        if key not in self.nodes:
            self.nodes[key] = {}
        assert title not in self.nodes[key]
        self.nodes[key][title] = (val, unit + " " + fmt_below(below), desc, sample)

    def flush(self):
        cpunames = sorted(self.cpunames)
        if not self.printed_header and not self.no_header:
            ts = ["Timestamp"] if self.timestamp else []
            header = ts + ["Area", "Node"] + cpunames + ["Description", "Sample", "Stddev", "Multiplex"]
            self.writer[self.curname].writerow([x for x in header])
            self.printed_header = True
        for key in sorted(sorted(self.nodes.keys(), key=lambda x: x[1]), key=lambda x: x[0] == ""):
            node = self.nodes[key]
            ts = [convert_ts(self.timestamp)] if self.timestamp else []
            l = ts + [key[0], key[1]]
            vlist = []
            ol = {}
            desc, sample = "", ""
            for cpuname in cpunames:
                if cpuname in node:
                    cpu = node[cpuname]
                    if cpu[2]:
                        desc = cpu[2]
                        desc = re.sub(r"\s+", " ", desc)
                    if cpu[3]:
                        sample = cpu[3]
                    # ignore unit for now
                    vlist.append(cpu[0])
                    ol[cpuname] = float(cpu[0].value) if cpu[0].value else ""
                else:
                    vlist.append(UVal("",0))
            l += [ol[x] if x in ol else "" for x in cpunames]
            l.append(desc)
            l.append(sample)
            vs = combine_uval(vlist)
            if vs:
                l += (vs.format_uncertainty().strip(), vs.format_mux().strip())
            else:
                l += ["", ""]
            self.writer[self.curname].writerow(l)
        self.nodes = {}

    print_footer = Output.print_footer_all

class OutputCSV(Output):
    """Output data in CSV format."""
    def __init__(self, logfile, sep, args, version, cpu):
        Output.__init__(self, logfile, version, cpu, args)
        self.writer = {}
        if self.logfiles:
            for n, f in self.logfiles.items():
                self.writer[n] = csv.writer(f, delimiter=sep, lineterminator='\n')
        else:
            self.writer[''] = csv.writer(self.logf, delimiter=sep, lineterminator='\n')
        self.args = args
        self.printed_headers = set()

    def print_header_csv(self, timestamp, title):
        if self.no_header:
            return
        if self.curname_nologf not in self.printed_headers:
            l = []
            if timestamp:
                l.append("Timestamp")
            if title:
                l.append("CPUs")
            self.writer[self.curname].writerow(l +
                ['Area', 'Value', 'Unit', 'Description',
                 'Sample', 'Stddev', 'Multiplex', 'Bottleneck', 'Idle'])
            self.printed_headers.add(self.curname_nologf)

    def show(self, timestamp, title, area, hdr, val, unit, desc, sample, bn, below, idle):
        self.print_header_csv(timestamp, title)
        if self.args.no_desc:
            desc = ""
        desc = re.sub(r"\s+", " ", desc)
        l = []
        if timestamp:
            l.append(convert_ts(timestamp))
        if title:
            l.append("CPU" + title if re.match(r'[0-9]+', title) else title)
        stddev = val.format_uncertainty().strip()
        multiplex = val.multiplex if not isnan(val.multiplex) else ""
        self.writer[self.curname].writerow(l + [hdr, val.format_value_raw().strip(),
                                  (unit + " " + fmt_below(below)).strip(),
                                  desc, sample, stddev, multiplex, bn, "Y" if idle else ""])

    print_footer = Output.print_footer_all

class OutputJSON(Output):
    """Output data in chrome / trace-viewer JSON format."""
    def __init__(self, logfile, sep, args, version, cpu):
        Output.__init__(self, logfile, version, cpu, args)
        self.nodes = defaultdict(dict) # type: DefaultDict[str, Dict[str, Any]]
        self.headers = OrderedDict()
        self.count = Counter() # type: typing.Counter[str]
        self.no_header = args.no_json_header
        self.no_footer = args.no_json_footer
        self.num = 0

    def print_footer_all(self):
        def write_all(s):
            if self.logfiles:
                for n in self.logfiles:
                    self.logfiles[n].write(s(n))
            else:
                self.logf.write(s(""))

        if self.no_footer:
            if self.num > 0:
                write_all(lambda x: ",\n")
        else:
            def start(name):
                n = ""
                if name not in self.count:
                    n += "[\n"
                return n + "\n]\n"
            write_all(start)

    print_footer = print_footer_all

    def show(self, timestamp, title, area, hdr, val, unit, desc, sample, bn, below, idle):
        self.timestamp = timestamp
        self.nodes[title][hdr] = val
        self.headers[hdr] = True
        self.num += 1

    def flush(self):
        nodes = OrderedDict() # type: OrderedDict[str,Any]
        for hdr in self.headers:
            for title in sorted(self.nodes.keys()):
                if hdr not in self.nodes[title]:
                    continue
                nd = self.nodes[title]
                val = nd[hdr].value if isinstance(nd[hdr], UVal) else nd[hdr]

                if title:
                    title += " "
                if hdr in ("Frontend_Bound", "Backend_Bound", "BadSpeculation", "Retiring"): # XXX
                    key = title + "Level1"
                    if key not in nodes:
                        nodes[key] = {}
                    nodes[title + "Level1"][hdr] = val
                elif hdr.count(".") >= 1:
                    dot = hdr.rindex(".")
                    nodes[title + hdr[:dot]] = { hdr: round(val, 2) }
                else: # assume it's metric
                    nodes[title + hdr] = {hdr: val}

        for name in nodes.keys():
            if self.count[self.curname] == 0:
                if not self.no_header:
                    self.logf.write("[\n")
            else:
                self.logf.write(",\n")
            json.dump({"name": name,
                      "ph": "C",
                      "pid": 0,
                      "ts": self.timestamp / 1e6 if self.timestamp and not isnan(self.timestamp) else 0,
                      "args": nodes[name]}, self.logf)
            self.count[self.curname] += 1
        self.nodes = defaultdict(dict)
        self.headers = OrderedDict()

    def remark(self, m):
        pass
